################################################################################
#
#  CDDL HEADER START
#
#  The contents of this file are subject to the terms of the Common Development
#  and Distribution License Version 1.0 (the "License").
#
#  You can obtain a copy of the license at
#  http:# www.opensource.org/licenses/CDDL-1.0.  See the License for the
#  specific language governing permissions and limitations under the License.
#
#  When distributing Covered Code, include this CDDL HEADER in each file and
#  include the License file in a prominent location with the name LICENSE.CDDL.
#  If applicable, add the following below this CDDL HEADER, with the fields
#  enclosed by brackets "[]" replaced with your own identifying information:
#
#  Portions Copyright (c) [yyyy] [name of copyright owner]. All rights reserved.
#
#  CDDL HEADER END
#
#  Copyright (c) 2017-2019, Regents of the University of Minnesota.
#  All rights reserved.
#
#  Contributor(s):
#     Ilia Nikiforov
#     Eric Fuemmler
#
################################################################################
"""
Helper classes for KIM Test Drivers

"""
import logging
import os
import shutil
from abc import ABC, abstractmethod
from copy import deepcopy
from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryDirectory
from typing import IO, Any, Dict, List, Optional, Union

import ase
import kim_edn
import numpy as np
import numpy.typing as npt
from ase import Atoms
from ase.calculators.calculator import Calculator
from ase.constraints import FixSymmetry
from ase.data import atomic_masses
from ase.filters import FrechetCellFilter, UnitCellFilter
from ase.optimize import LBFGSLineSearch
from ase.optimize.optimize import Optimizer
from kim_property import (
    get_properties,
    get_property_id_path,
    kim_property_create,
    kim_property_dump,
    kim_property_modify,
)
from kim_property.modify import STANDARD_KEYS_SCLAR_OR_WITH_EXTENT
from kim_query import raw_query

from ..aflow_util import (
    AFLOW,
    get_space_group_number_from_prototype,
    prototype_labels_are_equivalent,
)
from ..aflow_util.core import AFLOW_EXECUTABLE
from ..kimunits import convert_list, convert_units
from ..symmetry_util import (
    cartesian_rotation_is_in_point_group,
    change_of_basis_atoms,
    get_cell_from_poscar,
    get_change_of_basis_matrix_to_conventional_cell_from_formal_bravais_lattice,
    get_formal_bravais_lattice_from_space_group,
)

logger = logging.getLogger(__name__)
logging.basicConfig(filename="kim-tools.log", level=logging.INFO, force=True)

__author__ = ["ilia Nikiforov", "Eric Fuemmeler"]
__all__ = [
    "KIMTestDriverError",
    "KIMTestDriver",
    "get_crystal_structure_from_atoms",
    "get_poscar_from_crystal_structure",
    "get_atoms_from_crystal_structure",
    "SingleCrystalTestDriver",
    "query_crystal_structures",
    "detect_unique_crystal_structures",
    "get_deduplicated_property_instances",
    "minimize_wrapper",
]

# Force tolerance for the optional initial relaxation of the provided cell
FMAX_INITIAL = 1e-5
# Maximum steps for the optional initial relaxation of the provided cell
MAXSTEPS_INITIAL = 200
PROP_SEARCH_PATHS_INFO = (
    "- $KIM_PROPERTY_PATH (expanding globs including recursive **)\n"
    "- $PWD/local-props/**/\n"
    "- $PWD/local_props/**/"
)


def _get_optional_source_value(property_instance: Dict, key: str) -> Any:
    """
    Function for getting an optional key's source-value, or None if the key
    doesn't exist, from a Property Instance that is a Dict
    """
    if key in property_instance:
        return property_instance[key]["source-value"]
    else:
        return None


def _update_optional_key_in_property_dict(
    property_instance: Dict, key: str, value: Any, unit: Optional[str] = None
) -> None:
    """
    In a Property Instance that's a Dict, update a key or erase it if
    None is provided
    """
    if value is None:
        if key in property_instance:
            property_instance.pop(key)
    else:
        property_instance[key] = {"source-value": value}
        if unit is not None:
            property_instance[key]["source-unit"] = unit


def minimize_wrapper(
    atoms: Atoms,
    fmax: float = FMAX_INITIAL,
    steps: int = MAXSTEPS_INITIAL,
    variable_cell: bool = True,
    logfile: Optional[Union[str, IO]] = "kim-tools.log",
    algorithm: Optimizer = LBFGSLineSearch,
    cell_filter: UnitCellFilter = FrechetCellFilter,
    fix_symmetry: bool = False,
    opt_kwargs: Dict = {},
    flt_kwargs: Dict = {},
) -> bool:
    """
    Use LBFGSLineSearch (default) to Minimize cell energy with respect to cell shape and
    internal atom positions.

    LBFGSLineSearch convergence behavior is as follows:

    - The solver returns True if it is able to converge within the optimizer
      iteration limits (which can be changed by the ``steps`` argument passed
      to ``run``), otherwise it returns False.
    - The solver raises an exception in situations where the line search cannot
      improve the solution, typically due to an incompatibility between the
      potential's values for energy, forces, and/or stress.

    This routine attempts to minimizes the energy until the force and stress
    reduce below specified tolerances given a provided limit on the number of
    allowed steps. The code returns when convergence is achieved or no
    further progress can be made, either due to reaching the iteration step
    limit, or a stalled minimization due to line search failures.

    Args:
        atoms:
            Atomic configuration to be minimized.
        fmax:
            Force convergence tolerance (the magnitude of the force on each
            atom must be less than this for convergence)
        steps:
            Maximum number of iterations for the minimization
        variable_cell:
            True to allow relaxation with respect to cell shape
        logfile:
            Log file. ``'-'`` means STDOUT
        algorithm:
            ASE optimizer algorithm
        CellFilter:
            Filter to use if variable_cell is requested
        fix_symmetry:
            Whether to fix the crystallographic symmetry
        opt_kwargs:
            Dictionary of kwargs to pass to optimizer
        flt_kwargs:
            Dictionary of kwargs to pass to filter (e.g. "scalar_pressure")

    Returns:
        Whether the minimization succeeded
    """
    if fix_symmetry:
        symmetry = FixSymmetry(atoms)
        atoms.set_constraint(symmetry)
    if variable_cell:
        supercell_wrapped = cell_filter(atoms, **flt_kwargs)
        opt = algorithm(supercell_wrapped, logfile=logfile, **opt_kwargs)
    else:
        opt = algorithm(atoms, logfile=logfile, **opt_kwargs)
    try:
        converged = opt.run(fmax=fmax, steps=steps)
        iteration_limits_reached = not converged
        minimization_stalled = False
    except Exception as e:
        minimization_stalled = True
        iteration_limits_reached = False
        logger.info("The following exception was caught during minimization:")
        logger.info(repr(e))

    logger.info(
        "Minimization "
        + (
            "stalled"
            if minimization_stalled
            else "stopped" if iteration_limits_reached else "converged"
        )
        + " after "
        + (
            ("hitting the maximum of " + str(steps))
            if iteration_limits_reached
            else str(opt.nsteps)
        )
        + " steps."
    )

    del atoms.constraints

    if minimization_stalled or iteration_limits_reached:
        logger.info("Final forces:")
        logger.info(atoms.get_forces())
        logger.info("Final stress:")
        logger.info(atoms.get_stress())
        return False
    else:
        return True


def _add_property_instance(
    property_name: str,
    disclaimer: Optional[str] = None,
    property_instances: Optional[str] = None,
) -> str:
    """
    Initialize a new property instance in the serialized ``property_instances``. This
    wraps the ``kim_property.kim_property_create()`` function, with the simplification
    of setting "instance-id" automatically, as well as automatically searching for a
    definition file if ``property_name`` is not found.

    NOTE: Is there any need to allow Test Driver authors to specify an instance-id?

    Args:
        property_name:
            The property name, e.g.
            "tag:staff@noreply.openkim.org,2023-02-21:property/binding-energy-crystal"
            or "binding-energy-crystal"
        disclaimer:
            An optional disclaimer commenting on the applicability of this result, e.g.
            "This relaxation did not reach the desired tolerance."
        property_instances:
            A pre-existing EDN-serialized list of KIM Property instances to add to

    Returns:
            Updated EDN-serialized list of property instances
    """
    if property_instances is None:
        property_instances = "[]"
    # Get and check the instance-id to use.
    property_instances_deserialized = kim_edn.loads(property_instances)
    new_instance_index = len(property_instances_deserialized) + 1
    for property_instance in property_instances_deserialized:
        assert (
            property_instance["instance-id"] != new_instance_index
        ), "instance-id conflict"

    existing_properties = get_properties()
    property_in_existing_properties = False
    for existing_property in existing_properties:
        if (
            existing_property == property_name
            or get_property_id_path(existing_property)[3] == property_name
        ):
            property_in_existing_properties = True

    if not property_in_existing_properties:
        print(
            f"\nThe property name or id\n{property_name}\nwas not found in "
            "kim-properties.\n"
        )
        print(
            "I will now look for an .edn file containing its definition in the "
            f"following locations:\n{PROP_SEARCH_PATHS_INFO}\n"
        )

        property_search_paths = []

        # environment varible
        if "KIM_PROPERTY_PATH" in os.environ:
            property_search_paths += os.environ["KIM_PROPERTY_PATH"].split(":")

        # CWD
        property_search_paths.append(os.path.join(Path.cwd(), "local_props", "**"))
        property_search_paths.append(os.path.join(Path.cwd(), "local-props", "**"))

        # recursively search for .edn files in the paths, check if they are a property
        # definition with the correct name

        found_custom_property = False

        for search_path in property_search_paths:
            if found_custom_property:
                break
            else:
                # hack to expand globs in both absolute and relative paths
                if search_path[0] == "/":
                    base_path = Path("/")
                    search_glob = os.path.join(search_path[1:], "*.edn")
                else:
                    base_path = Path()
                    search_glob = os.path.join(search_path, "*.edn")

                for path in base_path.glob(search_glob):
                    if not os.path.isfile(
                        path
                    ):  # in case there's a directory named *.edn
                        continue
                    try:
                        path_str = str(path)
                        dict_from_edn = kim_edn.load(path_str)
                        if ("property-id") in dict_from_edn:
                            property_id = dict_from_edn["property-id"]
                            if (
                                property_id == property_name
                                or get_property_id_path(property_id)[3] == property_name
                            ):
                                property_name = path_str
                                found_custom_property = True
                                break
                    except Exception:
                        pass

        if not found_custom_property:
            raise KIMTestDriverError(
                f"\nThe property name or id\n{property_name}\nwas not found in "
                "kim-properties.\nI failed to find an .edn file containing a matching "
                '"property-id" key in the following locations:\n'
                + PROP_SEARCH_PATHS_INFO
            )

    return kim_property_create(
        new_instance_index, property_name, property_instances, disclaimer
    )


def _add_key_to_current_property_instance(
    property_instances: str,
    name: str,
    value: npt.ArrayLike,
    unit: Optional[str] = None,
    uncertainty_info: Optional[dict] = None,
) -> str:
    """
    Write a key to the last element of property_instances. This wraps
    ``kim_property.kim_property_modify()`` with a simplified (and more restricted)
    interface.

    Note: if the value is an array, this function will assume you want to write to the
    beginning of the array in every dimension.

    This function is intended to write entire keys in one go, and should not be used for
    modifying existing keys.

    WARNING! It is the developer's responsibility to make sure the array shape matches
    the extent specified in the property definition. This method uses fills the values
    of array keys as slices through the last dimension. If those slices are incomplete,
    kim_property automatically initializes the other elements in that slice to zero. For
    example, consider writing coordinates to a key with extent [":", 3]. The correct way
    to write a single atom would be to provide [[x, y, z]]. If you accidentally provide
    [[x], [y], [z]], it will fill the field with the coordinates
    [[x, 0, 0], [y, 0, 0], [z, 0, 0]]. This will not raise an error, only exceeding
    the allowed dimesnsions of the key will do so.

    Args:
        property_instances:
            An EDN-serialized list of dictionaries representing KIM Property Instances.
            The key will be added to the last dictionary in the list
        name:
            Name of the key, e.g. "cell-cauchy-stress"
        value:
            The value of the key. The function will attempt to convert it to a NumPy
            array, then use the dimensions of the resulting array. Scalars, lists,
            tuples, and arrays should work.
            Data type of the elements should be str, float, or int
        unit:
            The units
        uncertainty_info:
            dictionary containing any uncertainty keys you wish to include. See
            https://openkim.org/doc/schema/properties-framework/ for the possible
            uncertainty key names. These must be the same dimension as ``value``, or
            they may be scalars regardless of the shape of ``value``.

    Returns:
        Updated EDN-serialized list of property instances
    """

    def recur_dimensions(
        prev_indices: List[int],
        sub_value: npt.ArrayLike,
        modify_args: list,
        key_name: str = "source-value",
    ):
        sub_shape = sub_value.shape
        assert (
            len(sub_shape) != 0
        ), "Should not have gotten to zero dimensions in the recursive function"
        if len(sub_shape) == 1:
            # only if we have gotten to a 1-dimensional sub-array do we write stuff
            modify_args += [key_name, *prev_indices, "1:%d" % sub_shape[0], *sub_value]
        else:
            for i in range(sub_shape[0]):
                prev_indices.append(i + 1)  # convert to 1-based indices
                recur_dimensions(prev_indices, sub_value[i], modify_args, key_name)
                prev_indices.pop()

    value_arr = np.array(value)
    value_shape = value_arr.shape

    current_instance_index = len(kim_edn.loads(property_instances))
    modify_args = ["key", name]
    if len(value_shape) == 0:
        modify_args += ["source-value", value]
    else:
        prev_indices = []
        recur_dimensions(prev_indices, value_arr, modify_args)

    if unit is not None:
        modify_args += ["source-unit", unit]

    if uncertainty_info is not None:
        for uncertainty_key in uncertainty_info:
            if uncertainty_key not in STANDARD_KEYS_SCLAR_OR_WITH_EXTENT:
                raise KIMTestDriverError(
                    "Uncertainty key %s is not one of the allowed options %s."
                    % (uncertainty_key, str(STANDARD_KEYS_SCLAR_OR_WITH_EXTENT))
                )
            uncertainty_value = uncertainty_info[uncertainty_key]
            uncertainty_value_arr = np.array(uncertainty_value)
            uncertainty_value_shape = uncertainty_value_arr.shape

            if not (
                len(uncertainty_value_shape) == 0
                or uncertainty_value_shape == value_shape
            ):
                raise KIMTestDriverError(
                    f"The value {uncertainty_value_arr} provided for uncertainty key "
                    f"{uncertainty_key} has shape {uncertainty_value_shape}.\n"
                    f"It must either be a scalar or match the shape {value_shape} of "
                    "the source value you provided."
                )
            if len(uncertainty_value_shape) == 0:
                modify_args += [uncertainty_key, uncertainty_value]
            else:
                prev_indices = []
                recur_dimensions(
                    prev_indices, uncertainty_value_arr, modify_args, uncertainty_key
                )

    return kim_property_modify(property_instances, current_instance_index, *modify_args)


class KIMTestDriverError(Exception):
    def __init__(self, msg) -> None:
        # Call the base class constructor with the parameters it needs
        super(KIMTestDriverError, self).__init__(msg)
        self.msg = msg

    def __str__(self) -> str:
        return self.msg


class KIMTestDriver(ABC):
    """
    A base class for creating KIM Test Drivers. It has attributes that are likely
    to be useful to for most KIM tests

    Attributes:
        __kim_model_name (Optional[str]):
            KIM model name, absent if a non-KIM ASE calculator was provided
        __calc (:obj:`~ase.calculators.calculator.Calculator`):
            ASE calculator
        __output_property_instances (str):
            Property instances, possibly accumulated over multiple invocations of
            the Test Driver
    """

    def __init__(self, model: Union[str, Calculator]) -> None:
        """
        Args:
            model:
                ASE calculator or KIM model name to use
        """
        if isinstance(model, Calculator):
            self.__calc = model
            self.__kim_model_name = None
        else:
            from ase.calculators.kim.kim import KIM

            self.__kim_model_name = model
            self.__calc = KIM(self.__kim_model_name)

        self.__output_property_instances = "[]"

    def _setup(self, material, **kwargs) -> None:
        """
        Empty method, for optional overrides
        """
        pass

    @abstractmethod
    def _calculate(self, **kwargs) -> None:
        """
        Abstract calculate method
        """
        raise NotImplementedError("Subclasses must implement the _calculate method.")

    def __call__(self, material: Any = None, **kwargs) -> List[Dict]:
        """

        Main operation of a Test Driver:

            * Run :func:`~KIMTestDriver._setup` (the base class provides a barebones
              version, derived classes may override)
            * Call :func:`~KIMTestDriver._calculate` (implemented by each individual
              Test Driver)

        Args:
            material:
                Placeholder object for arguments describing the material to run
                the Test Driver on

        Returns:
            The property instances calculated during the current run
        """
        # count how many instances we had before we started
        previous_properties_end = len(kim_edn.loads(self.__output_property_instances))

        # _setup is likely overridden by an derived class
        self._setup(material, **kwargs)

        # implemented by each individual Test Driver
        self._calculate(**kwargs)

        # The current invocation returns a Python list of dictionaries containing all
        # properties computed during this run
        return kim_edn.loads(self.__output_property_instances)[previous_properties_end:]

    def _add_property_instance(
        self, property_name: str, disclaimer: Optional[str] = None
    ) -> None:
        """
        Initialize a new property instance.
        NOTE: Is there any need to allow Test Driver authors to specify an instance-id?

        Args:
            property_name:
                The property name, e.g.
                "tag:staff@noreply.openkim.org,2023-02-21:property/binding-energy-crystal"
                or "binding-energy-crystal"
            disclaimer:
                An optional disclaimer commenting on the applicability of this result,
                e.g. "This relaxation did not reach the desired tolerance."
        """
        self.__output_property_instances = _add_property_instance(
            property_name, disclaimer, self.__output_property_instances
        )

    def _add_key_to_current_property_instance(
        self,
        name: str,
        value: npt.ArrayLike,
        unit: Optional[str] = None,
        uncertainty_info: Optional[dict] = None,
    ) -> None:
        """
        Add a key to the most recent property instance added with
        :func:`~kim_tools.test_driver.core.KIMTestDriver._add_property_instance`.
        If the value is an
        array, this function will assume you want to write to the beginning of the array
        in every dimension. This function is intended to write entire keys in one go,
        and should not be used for modifying existing keys.

        WARNING! It is the developer's responsibility to make sure the array shape
        matches the extent specified in the property definition. This method uses
        ``kim_property.kim_property_modify``, and fills the values of array keys as
        slices through the last dimension. If those slices are incomplete, kim_property
        automatically initializes the other elements in that slice to zero. For example,
        consider writing coordinates to a key with extent [":", 3]. The correct way to
        write a single atom would be to provide [[x, y, z]]. If you accidentally provide
        [[x], [y], [z]], it will fill the field with the coordinates
        [[x, 0, 0], [y, 0, 0], [z, 0, 0]]. This will not raise an error, only exceeding
        the allowed dimesnsions of the key will do so.

        Args:
            name:
                Name of the key, e.g. "cell-cauchy-stress"
            value:
                The value of the key. The function will attempt to convert it to a NumPy
                array, then use the dimensions of the resulting array. Scalars, lists,
                tuples, and arrays should work. Data type of the elements should be str,
                float, int, or bool
            unit:
                The units
            uncertainty_info:
                dictionary containing any uncertainty keys you wish to include. See
                https://openkim.org/doc/schema/properties-framework/ for the possible
                uncertainty key names. These must be the same dimension as ``value``, or
                they may be scalars regardless of the shape of ``value``.
        """
        self.__output_property_instances = _add_key_to_current_property_instance(
            self.__output_property_instances, name, value, unit, uncertainty_info
        )

    def _add_file_to_current_property_instance(
        self, name: str, filename: os.PathLike, add_instance_id: bool = True
    ) -> None:
        """
        add a "file" type key-value pair to the current property instance.

        Args:
            name:
                Name of the key, e.g. "restart-file"
            filename:
                The path to the filename. If it is not in  "$CWD/output/",
                the file will be moved there
            add_instance_id:
                By default, a numerical index will be added before the file extension or
                at the end of a file with no extension. This is to ensure files do not
                get overwritten when the _calculate method is called repeatedly.

        Raises:
            KIMTestDriverError:
                If the provided filename does not exist
        """
        if not os.path.isfile(filename):
            raise KIMTestDriverError("Provided file {filename} does not exist.")

        # all paths here should be absolute
        cwd_path = Path(os.getcwd())
        output_path = cwd_path / "output"
        filename_path = Path(filename).resolve()

        if output_path not in filename_path.parents:
            # Need to move file to output
            if cwd_path in filename_path.parents:
                # The file is somewhere under CWD,
                # so move it under CWD/output with
                # its whole directory structure
                final_dir = os.path.join(
                    output_path, os.path.relpath(filename_path.parent)
                )
                os.makedirs(final_dir, exist_ok=True)
            else:
                # We got a file that isn't even under CWD.
                # I can't really hope to suss out what they
                # were going for, so just move it to CWD/output
                final_dir = output_path
        else:
            # already under output, not moving anything
            final_dir = filename_path.parent

        input_name = filename_path.name
        if add_instance_id:
            current_instance_id = len(kim_edn.loads(self.__output_property_instances))
            root, ext = os.path.splitext(input_name)
            root = root + "-" + str(current_instance_id)
            final_name = root + ext
        else:
            final_name = input_name

        final_path = os.path.join(final_dir, final_name)

        assert final_path.startswith(str(output_path))

        shutil.move(filename, final_path)

        # Filenames are reported relative to $CWD/output
        self._add_key_to_current_property_instance(
            name, os.path.relpath(final_path, output_path)
        )

    @property
    def kim_model_name(self) -> Optional[str]:
        """
        Get the KIM model name, if present
        """
        return self.__kim_model_name

    @property
    def property_instances(self) -> Dict:
        """
        Get all property instances accumulated over all calls to the Test Driver so far
        """
        return kim_edn.loads(self.__output_property_instances)

    @property
    def _calc(self) -> Optional[Calculator]:
        """
        Get the ASE calculator
        """
        return self.__calc

    def _get_serialized_property_instances(self) -> str:
        """
        Get the property instances computed so far in serialized EDN format
        """
        return self.__output_property_instances

    def _set_serialized_property_instances(self, property_instances: str) -> None:
        """
        Set the property instances from a serialized EDN string
        """
        self.__output_property_instances = property_instances

    def write_property_instances_to_file(self, filename="output/results.edn") -> None:
        """
        Write internal property instances (possibly accumulated over several calls to
        the Test Driver) to a file at the requested path. Also dumps any cached files to
        the same directory.

        Args:
            filename: path to write the file
        """

        with open(filename, "w") as f:
            kim_property_dump(
                self.__output_property_instances, f
            )  # serialize the dictionary to string first


def _add_common_crystal_genome_keys_to_current_property_instance(
    property_instances: str,
    prototype_label: str,
    stoichiometric_species: List[str],
    a: float,
    a_unit: str,
    parameter_values: Optional[List[float]] = None,
    library_prototype_label: Optional[Union[List[str], str]] = None,
    short_name: Optional[Union[List[str], str]] = None,
    cell_cauchy_stress: Optional[List[float]] = None,
    cell_cauchy_stress_unit: Optional[str] = None,
    temperature: Optional[float] = None,
    temperature_unit: Optional[str] = "K",
    crystal_genome_source_structure_id: Optional[List[List[str]]] = None,
    aflow_executable: str = AFLOW_EXECUTABLE,
) -> str:
    """
    Write common Crystal Genome keys to the last element of ``property_instances``. See
    https://openkim.org/properties/show/crystal-structure-npt for definition of the
    input keys. Note that the "parameter-names" key is inferred from the
    ``prototype_label`` input and is not an input to this function.

    Args:
        property_instances:
            An EDN-serialized list of dictionaries representing KIM Property Instances.
            The key will be added to the last dictionary in the list
        aflow_executable:
            Path to the AFLOW executable

    Returns:
        Updated EDN-serialized list of property instances
    """
    property_instances = _add_key_to_current_property_instance(
        property_instances, "prototype-label", prototype_label
    )
    property_instances = _add_key_to_current_property_instance(
        property_instances, "stoichiometric-species", stoichiometric_species
    )
    property_instances = _add_key_to_current_property_instance(
        property_instances, "a", a, a_unit
    )

    # get parameter names
    aflow = AFLOW(aflow_executable=aflow_executable)
    aflow_parameter_names = aflow.get_param_names_from_prototype(prototype_label)
    if parameter_values is None:
        if len(aflow_parameter_names) > 1:
            raise KIMTestDriverError(
                "The prototype label implies that parameter_values (i.e. dimensionless "
                "parameters besides a) are required, but you provided None"
            )
    else:
        if len(aflow_parameter_names) - 1 != len(parameter_values):
            raise KIMTestDriverError(
                "Incorrect number of parameter_values (i.e. dimensionless parameters "
                "besides a) for the provided prototype"
            )
        property_instances = _add_key_to_current_property_instance(
            property_instances, "parameter-names", aflow_parameter_names[1:]
        )
        property_instances = _add_key_to_current_property_instance(
            property_instances, "parameter-values", parameter_values
        )

    if short_name is not None:
        if not isinstance(short_name, list):
            short_name = [short_name]
        property_instances = _add_key_to_current_property_instance(
            property_instances, "short-name", short_name
        )

    if library_prototype_label is not None:
        property_instances = _add_key_to_current_property_instance(
            property_instances, "library-prototype-label", library_prototype_label
        )

    if cell_cauchy_stress is not None:
        if len(cell_cauchy_stress) != 6:
            raise KIMTestDriverError(
                "Please specify the Cauchy stress as a 6-dimensional vector in Voigt "
                "order [xx, yy, zz, yz, xz, xy]"
            )
        if cell_cauchy_stress_unit is None:
            raise KIMTestDriver("Please provide a `cell_cauchy_stress_unit`")
        property_instances = _add_key_to_current_property_instance(
            property_instances,
            "cell-cauchy-stress",
            cell_cauchy_stress,
            cell_cauchy_stress_unit,
        )

    if temperature is not None:
        if temperature_unit is None:
            raise KIMTestDriver("Please provide a `temperature_unit`")
        property_instances = _add_key_to_current_property_instance(
            property_instances, "temperature", temperature, temperature_unit
        )

    if crystal_genome_source_structure_id is not None:
        property_instances = _add_key_to_current_property_instance(
            property_instances,
            "crystal-genome-source-structure-id",
            crystal_genome_source_structure_id,
        )

    return property_instances


def _add_property_instance_and_common_crystal_genome_keys(
    property_name: str,
    prototype_label: str,
    stoichiometric_species: List[str],
    a: float,
    a_unit: str,
    parameter_values: Optional[List[float]] = None,
    library_prototype_label: Optional[Union[List[str], str]] = None,
    short_name: Optional[Union[List[str], str]] = None,
    cell_cauchy_stress: Optional[List[float]] = None,
    cell_cauchy_stress_unit: Optional[str] = None,
    temperature: Optional[float] = None,
    temperature_unit: Optional[str] = "K",
    crystal_genome_source_structure_id: Optional[List[List[str]]] = None,
    disclaimer: Optional[str] = None,
    property_instances: Optional[str] = None,
    aflow_executable: str = AFLOW_EXECUTABLE,
) -> str:
    """
    Initialize a new property instance to ``property_instances`` (an empty
    ``property_instances`` will be initialized if not provided). It will automatically
    get the an "instance-id" equal to the length of ``property_instances`` after it is
    added. Then, write the common Crystal Genome keys to it. See
    https://openkim.org/properties/show/crystal-structure-npt for definition of the
    input keys. Note that the "parameter-names" key is inferred from the
    ``prototype_label`` input and is not an input to this function.

    Args:
        property_name:
            The property name, e.g.
            "tag:staff@noreply.openkim.org,2023-02-21:property/binding-energy-crystal"
            or "binding-energy-crystal"
        disclaimer:
            An optional disclaimer commenting on the applicability of this result, e.g.
            "This relaxation did not reach the desired tolerance."
        property_instances:
            A pre-existing EDN-serialized list of KIM Property instances to add to
        aflow_executable:
            Path to the AFLOW executable

    Returns:
            Updated EDN-serialized list of property instances
    """
    property_instances = _add_property_instance(
        property_name, disclaimer, property_instances
    )
    return _add_common_crystal_genome_keys_to_current_property_instance(
        property_instances=property_instances,
        prototype_label=prototype_label,
        stoichiometric_species=stoichiometric_species,
        a=a,
        a_unit=a_unit,
        parameter_values=parameter_values,
        library_prototype_label=library_prototype_label,
        short_name=short_name,
        temperature=temperature,
        temperature_unit=temperature_unit,
        crystal_genome_source_structure_id=crystal_genome_source_structure_id,
        cell_cauchy_stress_unit=cell_cauchy_stress_unit,
        cell_cauchy_stress=cell_cauchy_stress,
        aflow_executable=aflow_executable,
    )


def get_crystal_structure_from_atoms(
    atoms: Atoms,
    get_short_name: bool = True,
    prim: bool = True,
    aflow_np: int = 4,
    aflow_executable: str = AFLOW_EXECUTABLE,
) -> Dict:
    """
    By performing a symmetry analysis on an :class:`~ase.Atoms` object, generate a
    dictionary that is a subset of the
    `crystal-structure-npt <https://openkim.org/properties/show/crystal-structure-npt>`_
    property. See https://openkim.org/doc/schema/properties-framework/ for more
    information about KIM properties and how values and units are defined. The
    dictionary returned by this function does not necessarily constitute a complete KIM
    Property Instance, but the key-value pairs are in valid KIM Property format and can
    be inserted as-is into any single crystal Crystal Genome property.

    Args:
        atoms: Configuration to analyze. It is assumed that the length unit is angstrom
        get_short_name:
            whether to compare against AFLOW prototype library to obtain short-name
        prim: whether to primitivize the atoms object first
        aflow_np: Number of processors to use with AFLOW executable
        aflow_executable:
            Path to the AFLOW executable

    Returns:
        A dictionary that has the following Property Keys (possibly optionally) defined.
        See the
        `crystal-structure-npt
        <https://openkim.org/properties/show/crystal-structure-npt>`_
        Property Definition for their meaning:

        - "stoichiometric-species"
        - "prototype-label"
        - "a"
        - "parameter-names" (inferred from "prototype-label")
        - "parameter-values"
        - "short-name"

    """
    aflow = AFLOW(aflow_executable=aflow_executable, np=aflow_np)

    proto_des = aflow.get_prototype_designation_from_atoms(atoms, prim=prim)
    library_prototype_label, short_name = (
        aflow.get_library_prototype_label_and_shortname_from_atoms(atoms, prim=prim)
        if get_short_name
        else (None, None)
    )

    a = proto_des["aflow_prototype_params_values"][0]
    parameter_values = (
        proto_des["aflow_prototype_params_values"][1:]
        if len(proto_des["aflow_prototype_params_values"]) > 1
        else None
    )

    property_instances = _add_property_instance_and_common_crystal_genome_keys(
        property_name="crystal-structure-npt",
        prototype_label=proto_des["aflow_prototype_label"],
        stoichiometric_species=sorted(list(set(atoms.get_chemical_symbols()))),
        a=a,
        a_unit="angstrom",
        parameter_values=parameter_values,
        library_prototype_label=library_prototype_label,
        short_name=short_name,
        aflow_executable=aflow_executable,
    )

    return kim_edn.loads(property_instances)[0]


def get_poscar_from_crystal_structure(
    crystal_structure: Dict,
    output_file: Optional[str] = None,
    flat: bool = False,
    addtl_args: str = "--webpage ",
    aflow_executable: str = AFLOW_EXECUTABLE,
) -> Optional[str]:
    """
    Write a POSCAR coordinate file (or output it as a multiline string) from the AFLOW
    Prototype Designation obtained from a KIM Property Instance. The following keys from
    https://openkim.org/properties/show/2023-02-21/staff@noreply.openkim.org/crystal-structure-npt
    are used:

    - "stoichiometric-species"
    - "prototype-label"
    - "a"
    - "parameter-values"
      (if "prototype-label" defines a crystal with free parameters besides "a")

    Note that although the keys are required to follow the schema of a KIM Property
    Instance (https://openkim.org/doc/schema/properties-framework/), there is no
    requirement or check that the input dictionary is an instance of any specific KIM
    Property, only that the keys required to build a crystal are present.

    Args:
        crystal_structure:
            Dictionary containing the required keys in KIM Property Instance format
        output_file:
            Name of the output file. If not provided, the output is returned as a string
        flat:
            whether the input dictionary is flattened
        addtl_args:
            additional arguments to --proto
        aflow_executable:
            path to AFLOW executable
    Returns:
        If ``output_file`` is not provided, a string in POSCAR format containg the
        primitive unit cell of the crystal as defined in
        http://doi.org/10.1016/j.commatsci.2017.01.017. Lengths are always in angstrom.
    Raises:
        AFLOW.ChangedSymmetryException:
            if the symmetry of the atoms object is different from ``prototype_label``
    """
    if flat:
        prototype_label = crystal_structure["prototype-label.source-value"]
        a_unit = crystal_structure["a.source-unit"]
        a_value = crystal_structure["a.source-value"]
        parameter_values_key = "parameter-values.source-value"
        stoichiometric_species = crystal_structure[
            "stoichiometric-species.source-value"
        ]
    else:
        prototype_label = crystal_structure["prototype-label"]["source-value"]
        a_unit = crystal_structure["a"]["source-unit"]
        a_value = crystal_structure["a"]["source-value"]
        parameter_values_key = "parameter-values"
        stoichiometric_species = crystal_structure["stoichiometric-species"][
            "source-value"
        ]

    aflow = AFLOW(aflow_executable=aflow_executable)
    aflow_parameter_names = aflow.get_param_names_from_prototype(prototype_label)

    # Atoms objects are always in angstrom
    if a_unit == "angstrom":
        a_angstrom = a_value
    else:
        a_angstrom = convert_units(
            a_value,
            crystal_structure["a"]["source-unit"],
            a_unit,
            True,
        )

    aflow_parameter_values = [a_angstrom]

    if parameter_values_key in crystal_structure:
        if len(aflow_parameter_names) == 1:
            raise KIMTestDriverError(
                f'Prototype label {prototype_label} implies only "a" parameter, but '
                'you provided "parameter-values"'
            )
        if flat:
            aflow_parameter_values += crystal_structure[parameter_values_key]
        else:
            aflow_parameter_values += crystal_structure[parameter_values_key][
                "source-value"
            ]

    try:
        return aflow.write_poscar_from_prototype(
            prototype_label=prototype_label,
            species=stoichiometric_species,
            parameter_values=aflow_parameter_values,
            output_file=output_file,
            addtl_args=addtl_args,
        )
    except AFLOW.ChangedSymmetryException as e:
        # re-raise, just indicating that this function knows about this exception
        raise e


def get_atoms_from_crystal_structure(
    crystal_structure: Dict,
    flat: bool = False,
    aflow_executable: str = AFLOW_EXECUTABLE,
) -> Atoms:
    """
    Generate an :class:`~ase.Atoms` object from the AFLOW Prototype Designation obtained
    from a KIM Property Instance. The following keys from
    https://openkim.org/properties/show/2023-02-21/staff@noreply.openkim.org/crystal-structure-npt
    are used:

    - "stoichiometric-species"
    - "prototype-label"
    - "a"
    - "parameter-values" (if "prototype-label" defines a crystal with free parameters
      besides "a")

    Note that although the keys are required to follow the schema of a KIM Property
    Instance (https://openkim.org/doc/schema/properties-framework/),
    There is no requirement or check that the input dictionary is an instance of any
    specific KIM Property, only that the keys required to build a crystal are present.

    Args:
        crystal_structure:
            Dictionary containing the required keys in KIM Property Instance format
        flat:
            whether the dictionary is flattened
        aflow_executable:
            path to AFLOW executable

    Returns:
        Primitive unit cell of the crystal as defined in the
        `AFLOW prototype standard <http://doi.org/10.1016/j.commatsci.2017.01.017>`_.
        Lengths are always in angstrom

    Raises:
        AFLOW.ChangedSymmetryException:
            if the symmetry of the atoms object is different from ``prototype_label``
    """
    try:
        poscar_string = get_poscar_from_crystal_structure(
            crystal_structure, flat=flat, aflow_executable=aflow_executable
        )
    except AFLOW.ChangedSymmetryException as e:
        # re-raise, just indicating that this function knows about this exception
        raise e

    with NamedTemporaryFile(mode="w+") as f:
        f.write(poscar_string)
        f.seek(0)
        atoms = ase.io.read(f.name, format="vasp")
    atoms.wrap()
    return atoms


class SingleCrystalTestDriver(KIMTestDriver):
    """
    A KIM test that computes property(s) of a single nominal crystal structure

    Attributes:
        __nominal_crystal_structure_npt [Dict]:
            An instance of the
            `crystal-structure-npt
            <https://openkim.org/properties/show/crystal-structure-npt>`_
            property representing the nominal crystal structure and conditions of the
            current call to the Test Driver.
        aflow_executable [str]:
            Path to the AFLOW executable
    """

    def __init__(
        self, model: Union[str, Calculator], aflow_executable: str = AFLOW_EXECUTABLE
    ) -> None:
        """
        Args:
            model:
                ASE calculator or KIM model name to use
            aflow_executable:
                Path to AFLOW executable
        """
        self.aflow_executable = aflow_executable
        super().__init__(model)

    def _setup(
        self,
        material: Union[Atoms, Dict],
        cell_cauchy_stress_eV_angstrom3: Optional[List[float]] = None,
        temperature_K: float = 0,
        **kwargs,
    ) -> None:
        """
        TODO: Consider allowing arbitrary units for temp and stress?

        Args:
            material:
                An :class:`~ase.Atoms` object or a KIM Property Instance specifying the
                nominal crystal structure that this run of the test
                will use. Pass one of the two following types of objects:

                Atoms object:

                    :class:`~ase.Atoms` object to use as the initial configuration. Note
                    that a symmetry analysis will be performed on it and a primitive
                    cell will be generated according to the conventions in
                    http://doi.org/10.1016/j.commatsci.2017.01.017. This primitive cell
                    may be rotated and translated relative to the configuration you
                    provided.

                Property instance:

                    Dictionary containing information about the nominal input crystal
                    structure in KIM Property Instance format (e.g. from a query to the
                    OpenKIM.org database)
                    The following keys from
                    https://openkim.org/properties/show/2023-02-21/staff@noreply.openkim.org/crystal-structure-npt
                    are used:

                    - "stoichiometric-species"
                    - "prototype-label"
                    - "a"
                    - "parameter-values"
                      (if "prototype-label" defines a crystal with
                      free parameters besides "a")
                    - "short-name" (if present)
                    - "crystal-genome-source-structure-id" (if present)

                    Note that although the keys are required to follow the schema of a
                    KIM Property Instance
                    (https://openkim.org/doc/schema/properties-framework/),
                    There is no requirement or check that the input dictionary is an
                    instance of any specific KIM Property, only that
                    the keys required to build a crystal are present.

            cell_cauchy_stress_eV_angstrom3:
                Cauchy stress on the cell in eV/angstrom^3 (ASE units) in
                [xx, yy, zz, yz, xz, xy] format. This is a nominal variable, and this
                class simply provides recordkeeping of it. It is up to derived classes
                to implement actually imposing this stress on the system.
            temperature_K:
                The temperature in Kelvin. This is a nominal variable, and this class
                simply provides recordkeeping of it. It is up to derived classes to
                implement actually setting the temperature of the system.
        """

        if cell_cauchy_stress_eV_angstrom3 is None:
            cell_cauchy_stress_eV_angstrom3 = [0, 0, 0, 0, 0, 0]

        if isinstance(material, Atoms):
            crystal_structure = get_crystal_structure_from_atoms(
                atoms=material, aflow_executable=self.aflow_executable
            )
            msg = (
                "Rebuilding atoms object in a standard setting defined by "
                "doi.org/10.1016/j.commatsci.2017.01.017. See log file or computed "
                "properties for the (possibly re-oriented) primitive cell that "
                "computations will be based on."
            )
            logger.info(msg)
            print()
            print(msg)
            print()
        else:
            crystal_structure = material

        # Pop the temperature and stress keys in case they came along with a query
        if "temperature" in crystal_structure:
            crystal_structure.pop("temperature")
        if "cell-cauchy-stress" in crystal_structure:
            crystal_structure.pop("cell-cauchy-stress")

        crystal_structure["temperature"] = {
            "source-value": temperature_K,
            "source-unit": "K",
        }
        crystal_structure["cell-cauchy-stress"] = {
            "source-value": cell_cauchy_stress_eV_angstrom3,
            "source-unit": "eV/angstrom^3",
        }
        if "meta" in crystal_structure:
            # Carrying 'meta' around doesn't really make sense since we may have already
            # modified things and will modify in the future
            crystal_structure.pop("meta")

        self.__nominal_crystal_structure_npt = crystal_structure

        # Warn if atoms appear unrelaxed
        atoms_tmp = self._get_atoms()
        force_max = np.max(atoms_tmp.get_forces())
        if force_max > FMAX_INITIAL:
            msg = (
                "The configuration you provided has a maximum force component "
                f"{force_max} eV/angstrom. Unless the Test Driver you are running "
                "provides minimization, you may wish to relax the configuration."
            )
            print(f"\nNOTE: {msg}\n")
            logger.info(msg)
        if cell_cauchy_stress_eV_angstrom3 != [0, 0, 0, 0, 0, 0]:
            stress_max = np.max(atoms_tmp.get_stress())
            if stress_max > FMAX_INITIAL:
                msg = (
                    "The configuration you provided has a maximum stress component "
                    f"{stress_max} eV/angstrom^3 even though the nominal state of the "
                    "system is unstressed. Unless the Test Driver you are running "
                    "provides minimization, you may wish to relax the configuration."
                )
                print(f"\nNOTE: {msg}\n")
                logger.info(msg)

    def _update_nominal_parameter_values(
        self,
        atoms: Atoms,
        max_resid: Optional[float] = None,
        cell_rtol: float = 0.01,
        rot_rtol: float = 0.01,
        rot_atol: float = 0.01,
    ) -> None:
        """
        Update the nominal parameter values of the nominal crystal structure from the
        provided :class:`~ase.Atoms` object. It is assumed that the crystallographic
        symmetry (space group + occupied Wyckoff positions) have not changed from the
        initially provided structure.

        The provided :class:`~ase.Atoms` object MUST be a primitive cell of the crystal
        as defined in http://doi.org/10.1016/j.commatsci.2017.01.017.
        The :class:`~ase.Atoms` object may be rotated, translated, and permuted, but the
        identity of the lattice vectors must be unchanged w.r.t. the crystallographic
        prototype. In other words, there must exist a permutation and translation of the
        fractional coordinates that enables them to match the equations defined by the
        prototype label.

        In practical terms, this means that this function is designed to take as input a
        relaxed or time-averaged from MD (and folded back into the original primitive
        cell) copy of the :class:`~ase.Atoms` object originally obtained from
        :func:`~kim_tools.test_driver.core.SingleCrystalTestDriver._get_atoms()`.

        If finding the parameter fails, this function will raise an exception. This
        probably indicates a phase transition to a different symmetry, which is a normal
        occasional occurrence if the original structure is not stable under the
        interatomic potential and prescribed conditions. These exceptions should not be
        handled and that run of the Test Driver should be allowed to fail.

        Args:
            atoms: Structure to analyze to get the new parameter values
            max_resid:
                Maximum residual allowed when attempting to match the fractional
                positions of the atoms to the crystallographic equations.
                If not provided, this is automatically set to 0.01*(minimum NN distance)
            cell_rtol:
                Relative tolerance on cell lengths and angles.
                Justification for default value: AFLOW uses 0.01*(minimum NN distance)
                as default tolerance.
            rot_rtol:
                Parameter to pass to :func:`numpy.allclose` for compariong fractional
                rotations. Default value chosen to be commensurate with AFLOW
                default distance tolerance of 0.01*(NN distance)
            rot_atol:
                Parameter to pass to :func:`numpy.allclose` for compariong fractional
                rotations. Default value chosen to be commensurate with AFLOW
                default distance tolerance of 0.01*(NN distance)

        Raises:
            AFLOW.FailedToMatchException:
                If the solution failed due to a failure to match two crystals that
                should be identical at some point in the solution process. This
                *usually* indicates a phase transformation
            AFLOW.ChangedSymmetryException:
                If a more definitive error indicating a phase transformation is
                encountered
        """
        aflow = AFLOW(aflow_executable=self.aflow_executable)
        try:
            (aflow_parameter_values, library_prototype_label, short_name) = (
                aflow.solve_for_params_of_known_prototype(
                    atoms=atoms,
                    prototype_label=self.__nominal_crystal_structure_npt[
                        "prototype-label"
                    ]["source-value"],
                    max_resid=max_resid,
                    cell_rtol=cell_rtol,
                    rot_rtol=rot_rtol,
                    rot_atol=rot_atol,
                )
            )
        except (AFLOW.FailedToMatchException, AFLOW.ChangedSymmetryException) as e:
            raise type(e)(
                "Encountered an error that MAY be the result of the nominal crystal "
                "being unstable under the given potential and conditions. Stopping "
                "execution."
            ) from e

        # Atoms objects always in angstrom
        self.__nominal_crystal_structure_npt["a"] = {
            "source-value": aflow_parameter_values[0],
            "source-unit": "angstrom",
        }
        if len(aflow_parameter_values) > 1:
            # If we detected new prototype values, old must have been
            # there and the same length as well
            old_param_values = self.__nominal_crystal_structure_npt["parameter-values"][
                "source-value"
            ]
            new_param_values = aflow_parameter_values[1:]
            assert len(old_param_values) == len(new_param_values)
            self.__nominal_crystal_structure_npt["parameter-values"] = {
                "source-value": new_param_values
            }
        _update_optional_key_in_property_dict(
            property_instance=self.__nominal_crystal_structure_npt,
            key="library-prototype-label",
            value=library_prototype_label,
        )
        _update_optional_key_in_property_dict(
            property_instance=self.__nominal_crystal_structure_npt,
            key="short-name",
            value=short_name,
        )

    def _verify_unchanged_symmetry(self, atoms: Atoms) -> bool:
        """
        Without changing the nominal state of the system, check if the provided Atoms
        object has the same symmetry as the nominal crystal structure associated with
        the current state of the Test Driver. This is defined as having the same
        prototype label, except for possible changes in Wyckoff letters as permitted by
        the space group normalizer.

        Args:
            atoms: The structure to compare to the current nominal structure of the Test
            Driver

        Returns:
            Whether or not the symmetry is unchanged

        """
        aflow = AFLOW(aflow_executable=self.aflow_executable)
        return prototype_labels_are_equivalent(
            aflow.get_prototype_designation_from_atoms(atoms)["aflow_prototype_label"],
            self.__nominal_crystal_structure_npt["prototype-label"]["source-value"],
        )

    def __add_poscar_to_curr_prop_inst(
        self,
        change_of_basis: Union[str, npt.ArrayLike],
        filename: os.PathLike,
        key_name: str,
    ) -> None:
        """
        Add a POSCAR file constructed from ``self.__nominal_crystal_structure_npt``
        to the current property instance.

        Args:
            change_of_basis:
                Passed to
                :meth:`kim_tools.test_driver.core.SingleCrystalTestDriver._get_atoms`
            filename:
                File to save to. Will be automatically moved and renamed,
                e.g. 'instance.poscar' -> 'output/instance-1.poscar'
            key_name:
                The property key to write to
        """

        # `_get_atoms` always returns in Angstrom
        atoms_tmp = self._get_atoms(change_of_basis)

        # will automatically be renamed
        # e.g. 'instance.poscar' -> 'output/instance-1.poscar'
        atoms_tmp.write(filename=filename, sort=True, format="vasp")
        self._add_file_to_current_property_instance(key_name, filename)

    def _add_property_instance_and_common_crystal_genome_keys(
        self,
        property_name: str,
        write_stress: Union[bool, List[float]] = False,
        write_temp: Union[bool, float] = False,
        stress_unit: Optional[str] = None,
        temp_unit: str = "K",
        disclaimer: Optional[str] = None,
    ) -> None:
        """
        Initialize a new property instance to ``self.property_instances``. It will
        automatically get the an "instance-id" equal to the length of
        ``self.property_instances`` after it is added. Then, write the common Crystal
        Genome keys to it from the attributes of this class.

        Args:
            property_name:
                The property name, e.g.
                "tag:staff@noreply.openkim.org,2023-02-21:property/binding-energy-crystal"
                or "binding-energy-crystal"
            write_stress:
                What (if any) to write to the "cell-cauchy-stress" key.
                If True, write the nominal stress the Test Driver was initialized with.
                If a list of floats is given, write that value (it must be a length 6
                list representing a stress tensor in Voigt order xx,yy,zz,yz,xz,xy).
                If a list is specified, you must also specify `stress_unit`.
            write_temp:
                What (if any) to write to the "temperature" key.
                If True, write the nominal temperature the Test Driver was initialized
                with. If float is given, write that value.
            stress_unit:
                Unit of stress. Required if a stress tensor is specified in
                `write_stress`.
            temp_unit:
                Unit of temperature. Defaults to K.
            disclaimer:
                An optional disclaimer commenting on the applicability of this result,
                e.g. "This relaxation did not reach the desired tolerance."
        """
        crystal_structure = self.__nominal_crystal_structure_npt

        a = crystal_structure["a"]["source-value"]

        a_unit = crystal_structure["a"]["source-unit"]

        prototype_label = crystal_structure["prototype-label"]["source-value"]

        stoichiometric_species = crystal_structure["stoichiometric-species"][
            "source-value"
        ]

        parameter_values = _get_optional_source_value(
            crystal_structure, "parameter-values"
        )

        library_prototype_label = _get_optional_source_value(
            crystal_structure, "library-prototype-label"
        )

        short_name = _get_optional_source_value(crystal_structure, "short-name")

        if write_stress is False:
            cell_cauchy_stress = None
            cell_cauchy_stress_unit = None
        elif write_stress is True:
            cell_cauchy_stress_stored = crystal_structure["cell-cauchy-stress"][
                "source-value"
            ]
            cell_cauchy_stress_unit_stored = crystal_structure["cell-cauchy-stress"][
                "source-unit"
            ]
            if (
                stress_unit is not None
                and stress_unit != cell_cauchy_stress_unit_stored
            ):
                cell_cauchy_stress_unit = stress_unit
                cell_cauchy_stress = convert_units(
                    cell_cauchy_stress_stored,
                    cell_cauchy_stress_unit_stored,
                    cell_cauchy_stress_unit,
                    True,
                )
            else:
                cell_cauchy_stress = cell_cauchy_stress_stored
                cell_cauchy_stress_unit = cell_cauchy_stress_unit_stored
        else:
            if len(write_stress) != 6:
                raise KIMTestDriverError(
                    "`write_stress` must be a boolean or an array of length 6"
                )
            cell_cauchy_stress = write_stress
            cell_cauchy_stress_unit = stress_unit

        if write_temp is False:
            temperature = None
            temperature_unit = None
        elif write_temp is True:
            temperature_stored = crystal_structure["temperature"]["source-value"]
            temperature_unit_stored = crystal_structure["temperature"]["source-unit"]
            if temp_unit != temperature_unit_stored:
                temperature_unit = temp_unit
                temperature = convert_units(
                    temperature_stored,
                    temperature_unit_stored,
                    temperature_unit,
                    True,
                )
            else:
                temperature = temperature_stored
                temperature_unit = temperature_unit_stored
        else:
            temperature = write_temp
            temperature_unit = temp_unit

        crystal_genome_source_structure_id = _get_optional_source_value(
            crystal_structure, "crystal-genome-source-structure-id"
        )

        super()._set_serialized_property_instances(
            _add_property_instance_and_common_crystal_genome_keys(
                property_name=property_name,
                prototype_label=prototype_label,
                stoichiometric_species=stoichiometric_species,
                a=a,
                a_unit=a_unit,
                parameter_values=parameter_values,
                library_prototype_label=library_prototype_label,
                short_name=short_name,
                cell_cauchy_stress=cell_cauchy_stress,
                cell_cauchy_stress_unit=cell_cauchy_stress_unit,
                temperature=temperature,
                temperature_unit=temperature_unit,
                crystal_genome_source_structure_id=crystal_genome_source_structure_id,
                disclaimer=disclaimer,
                property_instances=super()._get_serialized_property_instances(),
                aflow_executable=self.aflow_executable,
            )
        )

        self.__add_poscar_to_curr_prop_inst(
            "primitive", "instance.poscar", "coordinates-file"
        )
        self.__add_poscar_to_curr_prop_inst(
            "conventional",
            "conventional.instance.poscar",
            "coordinates-file-conventional",
        )

    def _get_temperature(self, unit: str = "K") -> float:
        """
        Get the nominal temperature

        Args:
            unit:
                The requested unit for the output. Must be understood by the GNU
                ``units`` utility
        """
        source_value = self.__nominal_crystal_structure_npt["temperature"][
            "source-value"
        ]
        source_unit = self.__nominal_crystal_structure_npt["temperature"]["source-unit"]
        if source_unit != unit:
            temp = convert_units(source_value, source_unit, unit, True)
        else:
            temp = source_value
        return temp

    def _get_cell_cauchy_stress(self, unit: str = "eV/angstrom^3") -> List[float]:
        """
        Get the nominal stress

        Args:
            unit:
                The requested unit for the output. Must be understood by the GNU
                ``units`` utility
        """
        source_value = self.__nominal_crystal_structure_npt["cell-cauchy-stress"][
            "source-value"
        ]
        source_unit = self.__nominal_crystal_structure_npt["cell-cauchy-stress"][
            "source-unit"
        ]
        if source_unit != unit:
            stress, _ = convert_list(source_value, source_unit, unit)
        else:
            stress = source_value
        return stress

    def _get_mass_density(self, unit: str = "amu/angstrom^3") -> float:
        """
        Get the mass density of the current nominal state of the system,
        according to the masses defined in :data:`ase.data.atomic_masses`

        Args:
            unit:
                The requested units

        Returns:
            The mass density of the crystal
        """
        atoms = self._get_atoms()  # always in angstrom
        vol_ang3 = atoms.get_volume()
        mass_amu = 0.0
        for atomic_number in atoms.get_atomic_numbers():
            mass_amu += atomic_masses[atomic_number]
        density_amu_ang3 = mass_amu / vol_ang3
        if unit != "amu/angstrom^3":
            density = convert_units(density_amu_ang3, "amu/angstrom^3", unit, True)
        else:
            density = density_amu_ang3

        return density

    def _get_nominal_crystal_structure_npt(self) -> Dict:
        """
        Get the dictionary returning the current nominal state of the system.

        Returns:
            An instance of the
            `crystal-structure-npt
            <https://openkim.org/properties/show/crystal-structure-npt>`_
            OpenKIM property containing a symmetry-reduced description of the nominal
            crystal structure.
        """
        return self.__nominal_crystal_structure_npt

    def deduplicate_property_instances(
        self,
        properties_to_deduplicate: Optional[List[str]] = None,
        allow_rotation: bool = False,
        aflow_np: int = 4,
        rot_rtol: float = 0.01,
        rot_atol: float = 0.01,
    ) -> None:
        """
        In the internally stored property instances,
        deduplicate any repeated crystal structures for each property id and merge
        their "crystal-genome-source-structure-id" keys.

        WARNING: Only the crystal structures are checked. If you for some reason have a
        property that can reasonably report different non-structural values for the
        same atomic configuration, this will delete the extras!

        Args:
            properties_to_deduplicate:
                A list of property names to pick out of ``property_instances`` to
                deduplicate. Each element can be the long or short name, e.g.
                "tag:staff@noreply.openkim.org,2023-02-21:property/binding-energy-crystal"
                or "binding-energy-crystal". If omitted, all properties will be
                deduplicated.
            allow_rotation:
                Whether or not structures that are rotated by a rotation that is not in
                the crystal's point group are considered identical
            aflow_np:
                Number of processors to use to run the AFLOW executable
            rot_rtol:
                Parameter to pass to :func:`numpy.allclose` for compariong fractional
                rotations. Default value chosen to be commensurate with AFLOW
                default distance tolerance of 0.01*(NN distance). Used only if
                `allow_rotation` is False
            rot_atol:
                Parameter to pass to :func:`numpy.allclose` for compariong fractional
                rotations. Default value chosen to be commensurate with AFLOW
                default distance tolerance of 0.01*(NN distance). Used only if
                `allow_rotation` is False
        """
        deduplicated_property_instances = get_deduplicated_property_instances(
            property_instances=self.property_instances,
            properties_to_deduplicate=properties_to_deduplicate,
            allow_rotation=allow_rotation,
            aflow_np=aflow_np,
            rot_rtol=rot_rtol,
            rot_atol=rot_atol,
            aflow_executable=self.aflow_executable,
        )
        logger.info(
            f"Deduplicated {len(self.property_instances)} Property Instances "
            f"down to {len(deduplicated_property_instances)}."
        )
        # Remove files
        for original_property_instance in self.property_instances:
            instance_id = original_property_instance["instance-id"]
            instance_id_still_there = False
            for deduplicated_property_instance in deduplicated_property_instances:
                if instance_id == deduplicated_property_instance["instance-id"]:
                    instance_id_still_there = True
                    break
            if not instance_id_still_there:
                for key in original_property_instance:
                    value_dict = original_property_instance[key]
                    if not isinstance(value_dict, dict):
                        continue
                    if "source-unit" in value_dict:
                        continue
                    value = value_dict["source-value"]
                    if isinstance(value, str):
                        # TODO: should really check the property def that this is a
                        # "file" type key, but come on, how incredibly unlikely is it
                        # that there just happends to be a string type property that
                        # happens to have a valid file path in it
                        candidate_filename = os.path.join("output", value)
                        if os.path.isfile(candidate_filename):
                            os.remove(candidate_filename)

        super()._set_serialized_property_instances(
            kim_edn.dumps(deduplicated_property_instances)
        )

    def _set_serialized_property_instances(self, property_instances) -> None:
        """
        An override to prevent SingleCrystalTestDriver derived classes from setting
        property instances directly
        """
        raise NotImplementedError(
            "Setting property instances directly not supported "
            "in Crystal Genome Test Drivers"
        )

    def _get_atoms(
        self, change_of_basis: Union[str, npt.ArrayLike] = "primitive"
    ) -> Atoms:
        """
        Get the atomic configuration representing the nominal crystal,
        with a calculator already attached.

        Args:
            change_of_basis:
                Specify the desired unit cell. The default, ``"primitive"``, gives
                the cell as defined in the `AFLOW prototype standard
                <http://doi.org/10.1016/j.commatsci.2017.01.017>`_. ``"conventional"``
                gives the conventional cell defined therein.

                Alternatively, provide an arbitrary change of basis matrix **P** as
                defined in ITA 1.5.1.2, with the above-defined primitive cell
                corresponding to the "old basis" and the returned ``Atoms`` object being
                in the "new basis".

                See the docstring for
                :func:`kim_tools.symmetry_util.core.change_of_basis_atoms` for
                more information on how to define the change of basis.

        Returns:
            Unit cell of the crystal.
            Lengths are always in angstrom
        """
        crystal_structure = self.__nominal_crystal_structure_npt
        atoms_prim = get_atoms_from_crystal_structure(
            crystal_structure, aflow_executable=self.aflow_executable
        )
        if isinstance(change_of_basis, str):
            if change_of_basis.lower() == "primitive":
                change_of_basis_matrix = None
            elif change_of_basis.lower() == "conventional":
                prototype_label = crystal_structure["prototype-label"]["source-value"]
                sgnum = get_space_group_number_from_prototype(prototype_label)
                formal_bravais_lattice = get_formal_bravais_lattice_from_space_group(
                    sgnum
                )
                change_of_basis_matrix = get_change_of_basis_matrix_to_conventional_cell_from_formal_bravais_lattice(  # noqa: E501
                    formal_bravais_lattice
                )
            else:
                raise KIMTestDriverError(
                    'Allowable string values for `change_of_basis` are "primitive" or '
                    f'"conventional". You provided f{change_of_basis}'
                )
        else:
            change_of_basis_matrix = change_of_basis

        if change_of_basis_matrix is None:
            atoms_tmp = atoms_prim
        else:
            atoms_tmp = change_of_basis_atoms(atoms_prim, change_of_basis_matrix)

        atoms_tmp.calc = self._calc
        return atoms_tmp


def query_crystal_structures(
    stoichiometric_species: List[str],
    prototype_label: Optional[str] = None,
    short_name: Optional[str] = None,
    cell_cauchy_stress_eV_angstrom3: Optional[List[float]] = None,
    temperature_K: float = 0,
    kim_model_name: Optional[str] = None,
) -> List[Dict]:
    """
    Query for all equilibrium parameter sets for this species combination and,
    optionally, crystal structure specified by ``prototype_label`` and/or ``short_name``
    in the KIM database. This is a utility function for running the test outside of the
    OpenKIM pipeline. In the OpenKIM pipeline, this information is delivered to the test
    driver through the ``runner`` script.

    Args:
        stoichiometric_species:
            List of unique species in the crystal. Required part of the Crystal Genome
            designation.
        short_name:
            short name of the crystal, e.g. "Hexagonal Close Packed". This will be
            searched as a case-insensitive regex, so partial matches will be returned.
            The list of possible shortnames is taken by postprocessing README_PROTO.TXT
            from the AFLOW software and packaged with kim-tools for reproducibility. To
            see the exact list of possible short names, call
            :func:`kim_tools.aflow_util.core.read_shortnames` and inspect the values of
            the returned
            dictionary. Note that a given short name corresponds to an exact set of
            parameters (with some tolerance), except the overall scale of the crystal.
            For example, "Hexagonal Close Packed" will return only structures with a
            c/a close to 1.63 (actually close packed), not any structure with the same
            symmetry as HCP as is sometimes colloquially understood.
            TODO: consider adding the same expanded logic we have on the website that
            searches for any crystal with the symmetry corresponding to a given
            shortname, i.e. invalidating the last caveat given above.
        prototype_label:
            AFLOW prototype label for the crystal.
        cell_cauchy_stress_eV_angstrom3:
            Cauchy stress on the cell in eV/angstrom^3 (ASE units) in
            [xx,yy,zz,yz,xz,xy] format
        temperature_K:
            The temperature in Kelvin
        kim_model_name:
            KIM model name. If not provided, RD will be queried instead

    Returns:
        List of kim property instances matching the query in the OpenKIM.org database
    """
    if cell_cauchy_stress_eV_angstrom3 is None:
        cell_cauchy_stress_eV_angstrom3 = [0, 0, 0, 0, 0, 0]

    stoichiometric_species.sort()

    # TODO: Some kind of generalized query interface for all tests, this is very
    # hand-made
    cell_cauchy_stress_Pa = [
        component * 1.6021766e11 for component in cell_cauchy_stress_eV_angstrom3
    ]

    query = {
        "meta.type": "rd" if kim_model_name is None else "tr",
        "property-id": (
            "tag:staff@noreply.openkim.org,2023-02-21:property/crystal-structure-npt"
        ),
        "stoichiometric-species.source-value": {
            "$size": len(stoichiometric_species),
            "$all": stoichiometric_species,
        },
        "cell-cauchy-stress.si-value": cell_cauchy_stress_Pa,
        "temperature.si-value": temperature_K,
    }

    if kim_model_name is not None:
        query["meta.subject.extended-id"] = kim_model_name

    if prototype_label is not None:
        query["prototype-label.source-value"] = prototype_label

    if short_name is not None:
        query["short-name.source-value"] = {"$regex": short_name, "$options": "$i"}

    raw_query_args = {
        "query": query,
        "database": "data",
        "limit": 0,
    }

    logger.info(f"Sending below query:\n{raw_query_args}")

    query_result = raw_query(**raw_query_args)

    len_msg = (
        f"Found {len(query_result)} equilibrium structures from "
        "query_crystal_genome_structures()"
    )
    logger.info(len_msg)
    logger.debug(f"Query result (length={len(query_result)}):\n{query_result}")

    print(f"!!! {len_msg} !!!")

    return query_result


def detect_unique_crystal_structures(
    crystal_structures: Union[List[Dict], Dict],
    allow_rotation: bool = False,
    aflow_np: int = 4,
    rot_rtol: float = 0.01,
    rot_atol: float = 0.01,
    aflow_executable=AFLOW_EXECUTABLE,
) -> Dict:
    """
    Detect which of the provided crystal structures is unique

    Args:
        crystal_structures:
            A list of dictionaries in KIM Property format, each containing the Crystal
            Genome keys required to build a structure, namely: "stoichiometric-species",
            "prototype-label", "a", and, if the prototype has free parameters,
            "parameter-values". These dictionaries are not required to be complete KIM
            Property Instances, e.g. the keys "property-id", "instance-id" and "meta"
            can be absent. Alternatively, this can be a dictionary of dictionaries with
            integers as indices, for the recursive call.
        allow_rotation:
            Whether or not structures that are rotated by a rotation that is not in the
            crystal's point group are considered identical
        aflow_np:
            Number of processors to use to run the AFLOW executable
        rot_rtol:
            Parameter to pass to :func:`numpy.allclose` for compariong fractional
            rotations. Default value chosen to be commensurate with AFLOW
            default distance tolerance of 0.01*(NN distance). Used only if
            `allow_rotation` is False
        rot_atol:
            Parameter to pass to :func:`numpy.allclose` for compariong fractional
            rotations. Default value chosen to be commensurate with AFLOW
            default distance tolerance of 0.01*(NN distance). Used only if
            `allow_rotation` is False
        aflow_executable:
            Path to AFLOW executable
    Returns:
        Dictionary with keys corresponding to indices of unique structures and values
        being lists of indices of their duplicates
    """
    if len(crystal_structures) == 0:
        return []

    aflow = AFLOW(aflow_executable=aflow_executable, np=aflow_np)

    with TemporaryDirectory() as tmpdirname:
        # I don't know if crystal_structurs is a list or a dict with integer keys
        for i in (
            range(len(crystal_structures))
            if isinstance(crystal_structures, list)
            else crystal_structures
        ):
            structure = crystal_structures[i]
            try:
                get_poscar_from_crystal_structure(
                    structure,
                    os.path.join(tmpdirname, str(i)),
                    aflow_executable=aflow_executable,
                )
            except AFLOW.ChangedSymmetryException:
                logger.info(
                    f"Comparison structure {i} failed to write a POSCAR due to a "
                    "detected higher symmetry"
                )

        comparison = aflow.compare_materials_dir(tmpdirname)

        unique_materials = {}
        for materials_group in comparison:
            i_repr = int(
                materials_group["structure_representative"]["name"].split("/")[-1]
            )
            unique_materials[i_repr] = []
            for structure_duplicate in materials_group["structures_duplicate"]:
                unique_materials[i_repr].append(
                    int(structure_duplicate["name"].split("/")[-1])
                )

        if not allow_rotation:
            for materials_group in comparison:
                # to preserve their ordering in the original input list, make this a
                # dictionary now
                repr_filename = materials_group["structure_representative"]["name"]
                rotated_structures = {}
                cell = get_cell_from_poscar(repr_filename)
                sgnum = materials_group["space_group"]
                for potential_rotated_duplicate in materials_group[
                    "structures_duplicate"
                ]:
                    cart_rot = potential_rotated_duplicate["rotation"]
                    if not cartesian_rotation_is_in_point_group(
                        cart_rot=cart_rot,
                        sgnum=sgnum,
                        cell=cell,
                        rtol=rot_rtol,
                        atol=rot_atol,
                    ):
                        i_rot_dup = int(
                            potential_rotated_duplicate["name"].split("/")[-1]
                        )
                        rotated_structures[i_rot_dup] = crystal_structures[i_rot_dup]
                        # Now that we know it's rotated, need to remove
                        # it from the list of duplicates
                        i_repr = int(repr_filename.split("/")[-1])
                        unique_materials[i_repr].remove(i_rot_dup)

                unique_materials.update(
                    detect_unique_crystal_structures(
                        crystal_structures=rotated_structures,
                        allow_rotation=False,
                        aflow_np=aflow_np,
                        rot_rtol=rot_rtol,
                        rot_atol=rot_atol,
                        aflow_executable=aflow_executable,
                    )
                )

    return unique_materials


def get_deduplicated_property_instances(
    property_instances: List[Dict],
    properties_to_deduplicate: Optional[List[str]] = None,
    allow_rotation: bool = False,
    aflow_np: int = 4,
    rot_rtol: float = 0.01,
    rot_atol: float = 0.01,
    aflow_executable: str = AFLOW_EXECUTABLE,
) -> List[Dict]:
    """
    Given a list of dictionaries constituting KIM Property instances,
    deduplicate any repeated crystal structures for each property id and merge
    their "crystal-genome-source-structure-id" keys.

    WARNING: Only the crystal structures are checked. If you for some reason have a
    property that can reasonably report different non-structural values for the
    same atomic configuration, this will delete the extras!

    Args:
        property_instances:
            The list of KIM Property Instances to deduplicate
        properties_to_deduplicate:
            A list of property names to pick out of ``property_instances`` to
            deduplicate. Each element can be the long or short name, e.g.
            "tag:staff@noreply.openkim.org,2023-02-21:property/binding-energy-crystal"
            or "binding-energy-crystal". If omitted, all properties will be
            deduplicated.
        allow_rotation:
            Whether or not structures that are rotated by a rotation that is not in the
            crystal's point group are considered identical
        aflow_np:
            Number of processors to use to run the AFLOW executable
        rot_rtol:
            Parameter to pass to :func:`numpy.allclose` for compariong fractional
            rotations. Default value chosen to be commensurate with AFLOW
            default distance tolerance of 0.01*(NN distance). Used only if
            `allow_rotation` is False
        rot_atol:
            Parameter to pass to :func:`numpy.allclose` for compariong fractional
            rotations. Default value chosen to be commensurate with AFLOW
            default distance tolerance of 0.01*(NN distance). Used only if
            `allow_rotation` is False
        aflow_executable:
            Path to aflow executable

    Returns:
        The deduplicated property instances
    """
    if properties_to_deduplicate is None:
        properties_set = set()
        for property_instance in property_instances:
            properties_set.add(property_instance["property-id"])
        properties_to_deduplicate = list(properties_set)

    property_instances_deduplicated = []
    for property_name in properties_to_deduplicate:
        # Pick out property instances with the relevant name
        property_instances_curr_name = []
        for property_instance in property_instances:
            property_id = property_instance["property-id"]
            if (
                property_id == property_name
                or get_property_id_path(property_id)[3] == property_name
            ):
                property_instances_curr_name.append(deepcopy(property_instance))
        if len(property_instances_curr_name) == 0:
            raise KIMTestDriverError(
                "The property you asked to deduplicate "
                "is not in the property instances you provided"
            )

        # Get unique-duplicate dictionary
        unique_crystal_structures = detect_unique_crystal_structures(
            crystal_structures=property_instances_curr_name,
            allow_rotation=allow_rotation,
            aflow_np=aflow_np,
            rot_rtol=rot_rtol,
            rot_atol=rot_atol,
            aflow_executable=aflow_executable,
        )

        # Put together the list of unique instances for the current
        # name only
        property_instances_curr_name_deduplicated = []
        for i_unique in unique_crystal_structures:
            property_instances_curr_name_deduplicated.append(
                property_instances_curr_name[i_unique]
            )
            # Put together a list of "crystal-genome-source-structure-id"
            # to gather into the deduplicated structure
            additional_source_structure_id = []
            for i_dup in unique_crystal_structures[i_unique]:
                source_structure_id = _get_optional_source_value(
                    property_instances_curr_name[i_dup],
                    "crystal-genome-source-structure-id",
                )
                if source_structure_id is not None:
                    # "crystal-genome-source-structure-id" is a 2D list
                    # but only the first row should be populated
                    additional_source_structure_id += source_structure_id[0]

            if len(additional_source_structure_id) != 0:
                if (
                    "crystal-genome-source-structure-id"
                    not in property_instances_curr_name_deduplicated[-1]
                ):
                    property_instances_curr_name_deduplicated[-1][
                        "crystal-genome-source-structure-id"
                    ] = [[]]
                property_instances_curr_name_deduplicated[-1][
                    "crystal-genome-source-structure-id"
                ]["source-value"][0] += additional_source_structure_id

        property_instances_deduplicated += property_instances_curr_name_deduplicated

    # Add any instances of properties that weren't deduplicated
    for property_instance in property_instances:
        property_id = property_instance["property-id"]
        if (
            property_id not in properties_to_deduplicate
            and get_property_id_path(property_id)[3] not in properties_to_deduplicate
        ):
            property_instances_deduplicated.append(deepcopy(property_instance))

    property_instances_deduplicated.sort(key=lambda a: a["instance-id"])

    return property_instances_deduplicated


# If called directly, do nothing
if __name__ == "__main__":
    pass



################################################################################
#
#  CDDL HEADER START
#
#  The contents of this file are subject to the terms of the Common Development
#  and Distribution License Version 1.0 (the "License").
#
#  You can obtain a copy of the license at
#  http:# www.opensource.org/licenses/CDDL-1.0.  See the License for the
#  specific language governing permissions and limitations under the License.
#
#  When distributing Covered Code, include this CDDL HEADER in each file and
#  include the License file in a prominent location with the name LICENSE.CDDL.
#  If applicable, add the following below this CDDL HEADER, with the fields
#  enclosed by brackets "[]" replaced with your own identifying information:
#
#  Portions Copyright (c) [yyyy] [name of copyright owner]. All rights reserved.
#
#  CDDL HEADER END
#
#  Copyright (c) 2017-2019, Regents of the University of Minnesota.
#  All rights reserved.
#
#  Contributor(s):
#     Ellad B. Tadmor
#     Daniel S. Karls
#
################################################################################
"""
Helper routines for KIM Tests and Verification Checks

"""

import itertools
import random

import numpy as np
from ase import Atoms
from ase.calculators.kim.kim import KIM
from ase.data import chemical_symbols

__all__ = [
    "KIMASEError",
    "atom_outside_cell_along_nonperiodic_dim",
    "check_if_atoms_interacting_energy",
    "check_if_atoms_interacting_force",
    "check_if_atoms_interacting",
    "get_isolated_energy_per_atom",
    "get_model_energy_cutoff",
    "fractional_coords_transformation",
    "perturb_until_all_forces_sizeable",
    "randomize_positions",
    "randomize_species",
    "remove_species_not_supported_by_ASE",
    "rescale_to_get_nonzero_energy",
    "rescale_to_get_nonzero_forces",
]


################################################################################
class KIMASEError(Exception):
    def __init__(self, msg):
        # Call the base class constructor with the parameters it needs
        super(KIMASEError, self).__init__(msg)
        self.msg = msg

    def __str__(self):
        return self.msg


################################################################################
def remove_species_not_supported_by_ASE(species):
    """
    Remove any species from the 'species' list that are not supported by ASE
    """
    supported_species = chemical_symbols[1:]
    return [s for s in species if s in supported_species]


################################################################################
def randomize_species(atoms, species, seed=None):
    """
    Given an ASE 'atoms' object, set random element for each atom selected
    from the list of available 'species' in a way that ensures all are
    represented with the same probabilities.
    """
    if seed is not None:
        random.seed(seed)

    # Indefinitely iterate through species putting them at random
    # unoccupied sites until all atoms are exhausted.
    indices = list(range(len(atoms)))
    num_occupied = 0
    for element in itertools.cycle(species):
        i = random.randint(0, len(indices) - 1)
        atoms[indices[i]].symbol = element
        del indices[i]
        num_occupied += 1
        if num_occupied == len(atoms):
            break


################################################################################
def fractional_coords_transformation(cell):
    """
    Given a set of cell vectors, this will return a transformation matrix T that can be
    used to multiply any arbitrary position to get its fractional coordinates in the
    basis of those cell vectors.
    """
    simulation_cell_volume = np.linalg.det(cell)

    T = (
        np.vstack(
            (
                np.cross(cell[:, 1], cell[:, 2]),
                np.cross(cell[:, 2], cell[:, 0]),
                np.cross(cell[:, 0], cell[:, 1]),
            )
        )
        / simulation_cell_volume
    )

    return T


################################################################################
def atom_outside_cell_along_nonperiodic_dim(T, atom_coord, pbc, tol=1e-12):
    """
    Given a transformation matrix to apply to an atomic position to get its fractional
    position in the basis of some cell vectors and the corresponding boundary
    conditions, determine if the atom is outside of the cell along any non-periodic
    directions (measured using tolerance 'tol').  This is relevant when using an SM from
    a simulator such as LAMMPS that performs spatial decomposition -- atoms that leave
    the box along non-periodic dimensions are likely to become "lost."
    """
    if all(pbc):
        # Skip the actual checks
        return False

    # Calculate fractional coordinate of this atom
    atom_coord_fractional = np.dot(T, atom_coord)
    for dof in range(0, 3):
        if not pbc[dof] and (
            atom_coord_fractional[dof] < -tol or atom_coord_fractional[dof] > 1 + tol
        ):
            return True

    # If we made it to here, this atom's position is OK
    return False


################################################################################
def randomize_positions(atoms, pert_amp, seed=None):
    """
    Given an ASE 'atoms' object, displace all atomic coordinates by a random amount in
    the range [-pert_amp, pert_amp] along *each* dimension.  Note that all atomic
    coordinates must be inside of the corresponding cell along non-periodic dimensions.
    As each atom is looped over, we continue generating perturbations until the
    displaced position is inside of the cell along any non-periodic directions
    (displacing outside the cell along periodic dimensions is allowed, although it's up
    to the calling function to wrap the positions if they need to be).
    """
    if seed is not None:
        random.seed(seed)

    pbc = atoms.get_pbc()
    if all(pbc):
        # Don't need to worry about moving atoms outside of cell
        for at in range(0, len(atoms)):
            atoms[at].position += [
                pert_amp * random.uniform(-1.0, 1.0) for i in range(3)
            ]

    else:
        # Get transformation matrix to get fractional coords
        T = fractional_coords_transformation(atoms.get_cell())

        for at in range(0, len(atoms)):
            # Check if the positional coordinate of this atom is valid to begin
            # with, i.e. it's inside the box along along all non-periodic directions
            if atom_outside_cell_along_nonperiodic_dim(T, atoms[at].position, pbc):
                raise KIMASEError(
                    "ERROR: Determined that atom {} with position {} is outside of the "
                    "simulation cell ({}) along one or more non-periodic directions.  "
                    "In order to prevent atoms from being lost, they must all be "
                    "contained inside of the simulation cell along non-periodic "
                    "dimensions.".format(at, atoms[at].position, atoms.get_cell())
                )

            for dof in range(0, 3):
                coord = atoms[at].position[dof].copy()
                done = False
                while not done:
                    atoms[at].position[dof] += random.uniform(-1.0, 1.0) * pert_amp
                    if not atom_outside_cell_along_nonperiodic_dim(
                        T, atoms[at].position, pbc
                    ):
                        done = True
                    else:
                        atoms[at].position[dof] = coord


################################################################################
def get_isolated_energy_per_atom(model, symbol):
    """
    Construct a non-periodic cell containing a single atom and compute its energy.
    """
    single_atom = Atoms(
        symbol,
        positions=[(0.1, 0.1, 0.1)],
        cell=(20, 20, 20),
        pbc=(False, False, False),
    )
    calc = KIM(model)
    single_atom.calc = calc
    energy_per_atom = single_atom.get_potential_energy()
    if hasattr(calc, "clean"):
        calc.clean()
    if hasattr(calc, "__del__"):
        calc.__del__()
    del single_atom
    return energy_per_atom


################################################################################
def rescale_to_get_nonzero_energy(atoms, isolated_energy_per_atom, etol):
    """
    If the given configuration has a potential energy, relative to the sum of the
    isolated energy corresponding to each atom present, smaller in magnitude than 'etol'
    (presumably because the distance between atoms is too large), rescale it making it
    smaller.  The 'isolated_energy_per_atom' arg should be a dict containing an entry
    for each atomic species present in the atoms object
    (additional entries are ignored).
    """
    num_atoms = len(atoms)
    if num_atoms < 2:
        # If we're only using a single atom, then we need to make sure that the cell is
        # periodic along at least one direction
        if num_atoms == 1:
            pbc = atoms.get_pbc()
            if not any(pbc):
                raise RuntimeError(
                    "ERROR: If only a single atom is present, the cell must "
                    "be periodic along at least one direction."
                )
        else:
            raise RuntimeError(
                "ERROR: Invalid configuration. Must have at least one atom"
            )

    if not isinstance(isolated_energy_per_atom, dict):
        raise ValueError(
            "Argument 'isolated_energy_per_atom' passed to "
            "rescale_to_get_nonzero_energy must be a dict containing and entry "
            "for each atomic species present in the atoms object."
        )

    # Check for any flat directions in the initial configuration and ignore them when
    # determining whether to stop rescaling
    pmin = atoms.get_positions().min(axis=0)  # minimum x,y,z coordinates
    pmax = atoms.get_positions().max(axis=0)  # maximum x,y,z coordinates
    delp = pmax - pmin  # system extent across x, y, z
    flat = [(extent <= np.finfo(extent).tiny) for extent in delp]

    # Compute the "trivial energy", i.e. the energy assuming none of the atoms interact
    # with each other at all
    species_of_each_atom = atoms.get_chemical_symbols()
    energy_trivial = 0.0
    for atom_species in species_of_each_atom:
        energy_trivial += isolated_energy_per_atom[atom_species]

    # Rescale cell and atoms
    cell = atoms.get_cell()
    energy = atoms.get_potential_energy()
    adjusted_energy = energy - energy_trivial
    if abs(adjusted_energy) <= etol:
        pmin = atoms.get_positions().min(axis=0)  # minimum x,y,z coordinates
        pmax = atoms.get_positions().max(axis=0)  # maximum x,y,z coordinates
        extent_along_nonflat_directions = [
            extent for direction, extent in enumerate(delp) if not flat[direction]
        ]
        delpmin = min(extent_along_nonflat_directions)
        while delpmin > np.finfo(delpmin).tiny:
            atoms.positions *= 0.5  # make configuration half the size
            cell *= 0.5  # make cell half the size
            atoms.set_cell(cell)  # need to adjust cell in case it's periodic
            delpmin *= 0.5
            energy = atoms.get_potential_energy()
            adjusted_energy = energy - energy_trivial
            if abs(adjusted_energy) > etol:
                return  # success!

        # Get species and write out error
        raise RuntimeError(
            "ERROR: Unable to scale configuration down to nonzero energy.  This was "
            "determined by computing the total potential energy relative to the sum of "
            "the isolated energy corresponding to each atom present and checking if "
            "the magnitude of the difference was larger than the supplied tolerance of "
            "{} eV.  This may mean that the species present in the cell ({}) do not "
            "have a non-trivial energy interaction for the  potential being used."
            "".format(etol, set(species_of_each_atom))
        )


################################################################################
def check_if_atoms_interacting_energy(model, symbols, etol):
    """
    First, get the energy of a single isolated atom of each species given in 'symbols'.
    Then, construct a dimer consisting of these two species and try to decrease its bond
    length until a discernible difference in the energy (from the sum of the isolated
    energy of each species) is detected.  The 'symbols' arg should be a list or tuple of
    length 2 indicating which species pair to check, e.g. to check if Al interacts with
    Al, one should specify ['Al', 'Al'].
    """
    if not isinstance(symbols, (list, tuple)) or len(symbols) != 2:
        raise ValueError(
            "Argument 'symbols' passed to check_if_atoms_interacting_energy "
            "must be a list of tuple of length 2 indicating the species pair to "
            "check"
        )

    isolated_energy_per_atom = {}
    isolated_energy_per_atom[symbols[0]] = get_isolated_energy_per_atom(
        model, symbols[0]
    )
    isolated_energy_per_atom[symbols[1]] = get_isolated_energy_per_atom(
        model, symbols[1]
    )

    dimer = Atoms(
        symbols,
        positions=[(0.1, 0.1, 0.1), (5.1, 0.1, 0.1)],
        cell=(20, 20, 20),
        pbc=(False, False, False),
    )
    calc = KIM(model)
    dimer.calc = calc
    try:
        rescale_to_get_nonzero_energy(dimer, isolated_energy_per_atom, etol)
        atoms_interacting = True
        return atoms_interacting
    except:  # noqa: E722
        atoms_interacting = False
        return atoms_interacting
    finally:
        if hasattr(calc, "clean"):
            calc.clean()
        if hasattr(calc, "__del__"):
            calc.__del__()
        del dimer


################################################################################
def check_if_atoms_interacting_force(model, symbols, ftol):
    """
    Construct a dimer and try to decrease its bond length until the force acting on each
    atom is larger than 'ftol' in magnitude.  The 'symbols' arg should be a list or
    tuple of length 2 indicating which species pair to check, e.g. to check if Al
    interacts with Al, one should specify ['Al', 'Al'].
    """
    if not isinstance(symbols, (list, tuple)) or len(symbols) != 2:
        raise ValueError(
            "Argument 'symbols' passed to check_if_atoms_interacting_force "
            "must be a list of tuple of length 2 indicating the species pair to "
            "check"
        )

    dimer = Atoms(
        symbols,
        positions=[(0.1, 0.1, 0.1), (5.1, 0.1, 0.1)],
        cell=(20, 20, 20),
        pbc=(False, False, False),
    )
    calc = KIM(model)
    dimer.calc = calc
    try:
        rescale_to_get_nonzero_forces(dimer, ftol)
        atoms_interacting = True
        return atoms_interacting
    except:  # noqa: E722
        atoms_interacting = False
        return atoms_interacting
    finally:
        if hasattr(calc, "clean"):
            calc.clean()
        if hasattr(calc, "__del__"):
            calc.__del__()
        del dimer


################################################################################
def check_if_atoms_interacting(
    model, symbols, check_energy=True, etol=1e-6, check_force=True, ftol=1e-3
):
    """
    Check to see whether non-trivial energy and/or forces can be detected using the
    current model.  The 'symbols' arg should be a list or tuple of length 2 indicating
    which species pair to check, e.g. to check if Al interacts with Al, one should
    specify ['Al', 'Al'].
    """
    if check_energy and not check_force:
        return check_if_atoms_interacting_energy(model, symbols, etol)
    elif not check_energy and check_force:
        return check_if_atoms_interacting_force(model, symbols, ftol)

    elif check_energy and check_force:
        atoms_interacting_energy = check_if_atoms_interacting_energy(
            model, symbols, etol
        )
        atoms_interacting_force = check_if_atoms_interacting_energy(
            model, symbols, ftol
        )
        return atoms_interacting_energy, atoms_interacting_force


################################################################################
def rescale_to_get_nonzero_forces(atoms, ftol):
    """
    If the given configuration has force components which are all smaller in absolute
    value than 'ftol' (presumably because the distance between atoms is too large),
    rescale it to be smaller until the largest force component in absolute value is
    greater than or equal to 'ftol'.  In a perfect crystal, the crystal is rescaled
    until the atoms on the surface reach the minimum value (internal atoms padded with
    another atoms around them will have zero force).  Note that any periodicity is
    turned off for the rescaling and then restored at the end.
    """
    if len(atoms) < 2:
        raise KIMASEError(
            "ERROR: Invalid configuration. Must have at least 2 atoms. Number of atoms "
            "= {}".format(len(atoms))
        )

    # Check for any flat directions in the initial configuration and ignore them when
    # determining whether to stop rescaling
    pmin = atoms.get_positions().min(axis=0)  # minimum x,y,z coordinates
    pmax = atoms.get_positions().max(axis=0)  # maximum x,y,z coordinates
    delp = pmax - pmin  # system extent across x, y, z
    flat = [(extent <= np.finfo(extent).tiny) for extent in delp]

    # Temporarily turn off any periodicity
    pbc_save = atoms.get_pbc()
    cell = atoms.get_cell()
    atoms.set_pbc([False, False, False])
    # Rescale cell and atoms
    forces = atoms.get_forces()
    if np.isnan(forces).any():
        raise RuntimeError("ERROR: Computed forces include at least one nan.")
    fmax = max(abs(forces.min()), abs(forces.max()))  # find max in abs value
    if fmax < ftol:
        pmin = atoms.get_positions().min(axis=0)  # minimum x,y,z coordinates
        pmax = atoms.get_positions().max(axis=0)  # maximum x,y,z coordinates
        extent_along_nonflat_directions = [
            extent for direction, extent in enumerate(delp) if not flat[direction]
        ]
        delpmin = min(extent_along_nonflat_directions)
        while delpmin > np.finfo(delpmin).tiny:
            atoms.positions *= 0.75  # make configuration 3/4 the size
            cell *= 0.75  # make cell 3/4 the size
            delpmin *= 0.75
            forces = atoms.get_forces()  # get max force
            fmax = max(abs(forces.min()), abs(forces.max()))
            if fmax >= ftol:
                # Restore periodicity
                atoms.set_pbc(pbc_save)
                atoms.set_cell(cell)
                return  # success!
        raise KIMASEError(
            "ERROR: Unable to scale configuration down to nonzero forces."
        )
    else:
        # Restore periodicity
        atoms.set_pbc(pbc_save)


################################################################################
def perturb_until_all_forces_sizeable(
    atoms, pert_amp, minfact=0.1, maxfact=5.0, max_iter=1000
):
    """
    Keep perturbing atoms in the ASE 'atoms' object until all force components on each
    atom have an absolute value of least 'minfact' times the largest (in absolute value)
    component across all force vectors coming in.  Note that all atomic coordinates must
    be inside of the corresponding cell along non-periodic dimensions.  Perturbations
    leading to a force component on any atom that is larger than 'maxfact' times the
    largest force component coming in are rejected.  Perturbations leading to atoms
    outside of the span across x, y, and z of the atomic positions coming in or outside
    of the simulation cell along non-periodic directions are also rejected.  The process
    repeats until max_iter iterations have been reached, at which point an exception is
    raised.
    """
    pbc = atoms.get_pbc()

    # Get transformation matrix to get fractional coords
    T = fractional_coords_transformation(atoms.get_cell())

    # First, ensure that all atomic positions are inside of the cell along non-periodic
    # dimensions
    if not all(pbc):
        for at in range(0, len(atoms)):
            # Check if the positional coordinate of this atom is valid to begin
            # with, i.e. it's inside the box along along all non-periodic directions
            if atom_outside_cell_along_nonperiodic_dim(T, atoms[at].position, pbc):
                raise KIMASEError(
                    "ERROR: Determined that atom {} with position {} is outside of the "
                    "simulation cell ({}) along one or more non-periodic directions.  "
                    "In order to prevent atoms from being lost, they must all be "
                    "contained inside of the simulation cell along non-periodic "
                    "dimensions.".format(at, atoms[at].position, atoms.get_cell())
                )

    forces = atoms.get_forces()
    if np.isnan(forces).any():
        raise RuntimeError("ERROR: Computed forces include at least one nan.")
    fmax = max(abs(forces.min()), abs(forces.max()))  # find max in abs value

    if fmax <= 1e2 * np.finfo(float).eps:
        raise KIMASEError(
            "ERROR: Largest force component on configuration is "
            "less than or equal to 1e2*machine epsilon. Cannot proceed."
        )

    pmin = atoms.get_positions().min(axis=0)  # minimum x,y,z coordinates
    pmax = atoms.get_positions().max(axis=0)  # maximum x,y,z coordinates
    saved_posns = atoms.get_positions().copy()
    saved_forces = atoms.get_forces().copy()
    some_forces_too_small = True

    # Counter to enforce max iterations
    iters = 0

    while some_forces_too_small:
        for at in range(0, len(atoms)):
            for dof in range(0, 3):
                if abs(forces[at, dof]) < minfact * fmax:
                    done = False
                    coord = atoms[at].position[dof].copy()
                    while not done:
                        atoms[at].position[dof] += random.uniform(-1.0, 1.0) * pert_amp
                        if (
                            pmin[dof] <= atoms[at].position[dof] <= pmax[dof]
                        ) and not atom_outside_cell_along_nonperiodic_dim(
                            T, atoms[at].position, pbc
                        ):
                            done = True
                        else:
                            atoms[at].position[dof] = coord
        try:
            forces = atoms.get_forces()
            if np.isnan(forces).any():
                raise RuntimeError("ERROR: Computed forces include at least one nan.")
            fmax_new = max(abs(forces.min()), abs(forces.max()))
            if fmax_new > maxfact * fmax:
                # forces too large, abort perturbation
                atoms.set_positions(saved_posns)
                forces = saved_forces.copy()
                continue
            fmin_new = min(abs(forces.min()), abs(forces.max()))
            if fmin_new > minfact * fmax:
                some_forces_too_small = False
        except:  # noqa: E722
            # force calculation failed, abort perturbation
            atoms.set_positions(saved_posns)
            continue
        finally:
            iters = iters + 1
            if iters == max_iter:
                raise KIMASEError(
                    "Maximum iterations ({}) exceeded in call to "
                    "function perturb_until_all_forces_sizeable()".format(max_iter)
                )


################################################################################
def get_model_energy_cutoff(
    model,
    symbols,
    xtol=1e-8,
    etol_coarse=1e-6,
    etol_fine=1e-15,
    max_bisect_iters=1000,
    max_upper_cutoff_bracket=20.0,
):
    """
    Compute the distance at which energy interactions become non-trival for a given
    model and a species pair it supports.  This is done by constructing a dimer composed
    of these species in a large finite box, increasing the separation if necessary until
    the total potential energy is within 'etol_fine' of the sum of the corresponding
    isolated energies, and then shrinking the separation until the energy differs from
    that value by more than 'etol_coarse'.  Using these two separations to bound the
    search range, bisection is used to refine in order to locate the cutoff.  The
    'symbols' arg should be a list or tuple of length 2 indicating which species pair to
    check, e.g. to get the energy cutoff of Al with Al, one should specify ['Al', 'Al'].

    This function is based on the content of the DimerContinuityC1__VC_303890932454_002
    Verification Check in OpenKIM [1-3].

    [1] Tadmor E. Verification Check of Dimer C1 Continuity v002. OpenKIM; 2018.
        doi:10.25950/43d2c6d5

    [2] Tadmor EB, Elliott RS, Sethna JP, Miller RE, Becker CA. The potential of
        atomistic simulations and the Knowledgebase of Interatomic Models. JOM.
        2011;63(7):17. doi:10.1007/s11837-011-0102-6

    [3] Elliott RS, Tadmor EB. Knowledgebase of Interatomic Models (KIM) Application
        Programming Interface (API). OpenKIM; 2011. doi:10.25950/ff8f563a
    """
    from scipy.optimize import bisect

    def get_dimer_positions(a, large_cell_len):
        """
        Generate positions for a dimer of length 'a' centered in a finite simulation box
        with side length 'large_cell_len'
        """
        half_cell = 0.5 * large_cell_len
        positions = [
            [half_cell - 0.5 * a, half_cell, half_cell],
            [half_cell + 0.5 * a, half_cell, half_cell],
        ]
        return positions

    def energy(a, dimer, large_cell_len, einf):
        dimer.set_positions(get_dimer_positions(a, large_cell_len))
        return dimer.get_potential_energy() - einf

    def energy_cheat(a, dimer, large_cell_len, offset, einf):
        dimer.set_positions(get_dimer_positions(a, large_cell_len))
        return (dimer.get_potential_energy() - einf) + offset

    if not isinstance(symbols, (list, tuple)) or len(symbols) != 2:
        raise ValueError(
            "Argument 'symbols' passed to check_if_atoms_interacting_energy "
            "must be a list of tuple of length 2 indicating the species pair to "
            "check"
        )

    isolated_energy_per_atom = {}
    isolated_energy_per_atom[symbols[0]] = get_isolated_energy_per_atom(
        model, symbols[0]
    )
    isolated_energy_per_atom[symbols[1]] = get_isolated_energy_per_atom(
        model, symbols[1]
    )
    einf = isolated_energy_per_atom[symbols[0]] + isolated_energy_per_atom[symbols[1]]

    # First, establish the upper bracket cutoff by starting at 'b_init' Angstroms and
    # incrementing by 'db' until
    b_init = 4.0

    # Create finite box of size large_cell_len
    large_cell_len = 50
    dimer = Atoms(
        symbols,
        positions=get_dimer_positions(b_init, large_cell_len),
        cell=(large_cell_len, large_cell_len, large_cell_len),
        pbc=(False, False, False),
    )
    calc = KIM(model)
    dimer.calc = calc

    db = 2.0
    b = b_init - db
    still_interacting = True
    while still_interacting:
        b += db
        if b > max_upper_cutoff_bracket:
            if hasattr(calc, "clean"):
                calc.clean()
            if hasattr(calc, "__del__"):
                calc.__del__()

            raise KIMASEError(
                "Exceeded limit on upper bracket when determining cutoff "
                "search range"
            )
        else:
            eb = energy(b, dimer, large_cell_len, einf)
            if abs(eb) < etol_fine:
                still_interacting = False

    a = b
    da = 0.01
    not_interacting = True
    while not_interacting:
        a -= da
        if a < 0:
            if hasattr(calc, "clean"):
                calc.clean()
            if hasattr(calc, "__del__"):
                calc.__del__()

            raise RuntimeError(
                "Failed to determine lower bracket for cutoff search using etol_coarse "
                "= {}.  This may mean that the species pair provided ({}) does not "
                "have a non-trivial energy interaction for the potential being "
                "used.".format(etol_coarse, symbols)
            )
        else:
            ea = energy(a, dimer, large_cell_len, einf)
            if abs(ea) > etol_coarse:
                not_interacting = False

    # NOTE: Some Simulator Models have a history dependence due to them maintaining
    #       charges from the previous energy evaluation to use as an initial guess
    #       for the next charge equilibration.  We therefore have to treat them not
    #       as single-valued functions but as distributions, i.e.  for a given
    #       configuration you might get any of a range of energy values depending on
    #       the history of your previous energy evaluations.  This is particularly
    #       problematic for this step, where we set up a bisection problem in order
    #       to determine the cutoff radius of the model.  Our solution for this
    #       specific case is to make a very crude estimate of the variance of that
    #       distribution with a 10% factor of safety on it.
    eb_new = energy(b, dimer, large_cell_len, einf)
    eb_error = abs(eb_new - eb)

    # compute offset to ensure that energy before and after cutoff have
    # different signs
    if ea < eb:
        offset = -eb + 1.1 * eb_error + np.finfo(float).eps
    else:
        offset = -eb - 1.1 * eb_error - np.finfo(float).eps

    rcut, results = bisect(
        energy_cheat,
        a,
        b,
        args=(dimer, large_cell_len, offset, einf),
        full_output=True,
        xtol=xtol,
        maxiter=max_bisect_iters,
    )

    # General clean-up
    if hasattr(calc, "clean"):
        calc.clean()
    if hasattr(calc, "__del__"):
        calc.__del__()

    if not results.converged:
        raise RuntimeError(
            "Bisection search to find cutoff distance did not converge "
            "within {} iterations with xtol = {}".format(max_bisect_iters, xtol)
        )
    else:
        return rcut


# If called directly, do nothing
if __name__ == "__main__":
    pass



__version__ = "0.3.1"

from .aflow_util import *
from .aflow_util import __all__ as aflow_all
from .ase import *
from .ase import __all__ as ase_all
from .symmetry_util import *
from .symmetry_util import __all__ as symmetry_all
from .test_driver import *
from .test_driver import __all__ as test_driver_all
from .vc import *
from .vc import __all__ as vc_all

__all__ = test_driver_all + aflow_all + symmetry_all + ase_all + vc_all



"""
Crystal Symmetry utilities and data that are (mostly) independent of AFLOW
"""

import json
import logging
import os
from itertools import product
from math import ceil
from typing import Dict, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
from ase import Atoms
from ase.cell import Cell
from ase.geometry import get_distances, get_duplicate_atoms
from matplotlib.backends.backend_pdf import PdfPages
from pymatgen.core.operations import SymmOp
from pymatgen.core.tensors import Tensor
from scipy.stats import kstest
from sklearn.decomposition import PCA
from sympy import Matrix, cos, matrix2numpy, sin, sqrt, symbols

logger = logging.getLogger(__name__)
logging.basicConfig(filename="kim-tools.log", level=logging.INFO, force=True)

__all__ = [
    "BRAVAIS_LATTICES",
    "FORMAL_BRAVAIS_LATTICES",
    "CENTERING_DIVISORS",
    "C_CENTERED_ORTHORHOMBIC_GROUPS",
    "A_CENTERED_ORTHORHOMBIC_GROUPS",
    "IncorrectCrystallographyException",
    "IncorrectNumAtomsException",
    "are_in_same_wyckoff_set",
    "space_group_numbers_are_enantiomorphic",
    "cartesian_to_fractional_itc_rotation_from_ase_cell",
    "cartesian_rotation_is_in_point_group",
    "get_cell_from_poscar",
    "get_wyck_pos_xform_under_normalizer",
    "get_bravais_lattice_from_space_group",
    "get_formal_bravais_lattice_from_space_group",
    "get_primitive_wyckoff_multiplicity",
    "get_symbolic_cell_from_formal_bravais_lattice",
    "get_change_of_basis_matrix_to_conventional_cell_from_formal_bravais_lattice",
    "change_of_basis_atoms",
    "get_possible_primitive_shifts",
    "get_primitive_genpos_ops",
]

C_CENTERED_ORTHORHOMBIC_GROUPS = (20, 21, 35, 36, 37, 63, 64, 65, 66, 67, 68)
A_CENTERED_ORTHORHOMBIC_GROUPS = (38, 39, 40, 41)
BRAVAIS_LATTICES = [
    "aP",
    "mP",
    "mC",
    "oP",
    "oC",
    "oI",
    "oF",
    "tP",
    "tI",
    "hP",
    "hR",
    "cP",
    "cF",
    "cI",
]
FORMAL_BRAVAIS_LATTICES = BRAVAIS_LATTICES + ["oA"]
CENTERING_DIVISORS = {
    "P": 1,
    "C": 2,
    "A": 2,
    "I": 2,
    "F": 4,
    "R": 3,
}
DATA_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), "data")


class IncorrectCrystallographyException(Exception):
    """
    Raised when incorrect data is provided, e.g. nonexistent Bravais lattice etc.
    """


class IncorrectNumAtomsException(Exception):
    """
    Raised when the a disagreement in the number of atoms is found.
    """


class PeriodExtensionException(Exception):
    """
    Raised when a period-extending phase transition is detected.
    """


def _check_space_group(sgnum: Union[int, str]):
    try:
        assert 1 <= int(sgnum) <= 230
    except Exception:
        raise IncorrectCrystallographyException(
            f"Got a space group number {sgnum} that is non-numeric or not between 1 "
            "and 230 inclusive"
        )


def cartesian_to_fractional_itc_rotation_from_ase_cell(
    cart_rot: npt.ArrayLike, cell: npt.ArrayLike
) -> npt.ArrayLike:
    """
    Convert Cartesian to fractional rotation. Read the arguments and returns carefully,
    as there is some unfortunate mixing of row and columns because of the different
    conventions of the ITC and ASE and other simulation packages

    Args:
        cart_rot:
            Cartesian rotation. It is assumed that this is for left-multiplying column
            vectors, although in cases where we don't care if we're working with the
            rotation or its inverse (e.g. when checking whether or not it's in the
            point group), this doesn't matter due to orthogonality
        cell:
            The cell of the crystal, with each row being a cartesian vector
            representing a lattice vector. This is consistent with most simulation
            packages, but transposed from the ITC

    Returns:
        The fractional rotation in ITC convention, i.e. for left-multiplying column
        vectors. Here the distinction with a matrix's transpose DOES matter, because
        the fractional coordinate system is not orthonormal.
    """

    cell_arr = np.asarray(cell)
    cart_rot_arr = np.asarray(cart_rot)

    if not ((cell_arr.shape == (3, 3)) and (cart_rot_arr.shape == (3, 3))):
        raise IncorrectCrystallographyException(
            "Either the rotation matrix or the cell provided were not 3x3 matrices"
        )

    return np.transpose(cell_arr @ cart_rot_arr.T @ np.linalg.inv(cell_arr))


def fractional_to_cartesian_itc_rotation_from_ase_cell(
    frac_rot: npt.ArrayLike, cell: npt.ArrayLike
) -> npt.ArrayLike:
    """
    Convert fractional to Cartesian rotation. Read the arguments and returns carefully,
    as there is some unfortunate mixing of row and columns because of the different
    conventions of the ITC and ASE and other simulation packages

    Args:
        frac_rot:
            The fractional rotation in ITC convention, i.e. for left-multiplying column
            vectors. Here the distinction with a matrix's transpose DOES matter, because
            the fractional coordinate system is not orthonormal.
        cell:
            The cell of the crystal, with each row being a cartesian vector
            representing a lattice vector. This is consistent with most simulation
            packages, but transposed from the ITC

    Returns:
        Cartesian rotation. It is assumed that this is for left-multiplying column
        vectors, although in cases where we don't care if we're working with the
        rotation or its inverse (e.g. when checking whether or not it's in the
        point group), this doesn't matter due to orthogonality
    """

    cell_arr = np.asarray(cell)
    frac_rot_arr = np.asarray(frac_rot)

    if not ((cell_arr.shape == (3, 3)) and (frac_rot_arr.shape == (3, 3))):
        raise IncorrectCrystallographyException(
            "Either the rotation matrix or the cell provided were not 3x3 matrices"
        )

    return np.transpose(np.linalg.inv(cell_arr) @ frac_rot_arr.T @ cell_arr)


def cartesian_rotation_is_in_point_group(
    cart_rot: npt.ArrayLike,
    sgnum: Union[int, str],
    cell: npt.ArrayLike,
    rtol: float = 1e-2,
    atol: float = 1e-2,
) -> bool:
    """
    Check that a Cartesian rotation is in the point group of a crystal given by its
    space group number and primitive cell

    Args:
        cart_rot:
            Cartesian rotation
        sgnum:
            space group number
        cell:
            The *primitive* cell of the crystal as defined in
            http://doi.org/10.1016/j.commatsci.2017.01.017, with each row being a
            cartesian vector representing a lattice vector. This is
            consistent with most simulation packages, but transposed from the ITC
        rtol:
            Parameter to pass to :func:`numpy.allclose` for compariong fractional
            rotations. Default value chosen to be commensurate with AFLOW
            default distance tolerance of 0.01*(NN distance)
        atol:
            Parameter to pass to :func:`numpy.allclose` for compariong fractional
            rotations. Default value chosen to be commensurate with AFLOW
            default distance tolerance of 0.01*(NN distance)
    """
    # we don't care about properly transposing (i.e. worrying whether it's operating on
    # row or column vectors) the input cart_rot because that one is orthogonal, and
    # both it and its inverse must be in the point group
    frac_rot = cartesian_to_fractional_itc_rotation_from_ase_cell(cart_rot, cell)

    space_group_ops = get_primitive_genpos_ops(sgnum)

    logger.info(f"Attempting to match fractional rotation:\n{frac_rot}")

    for op in space_group_ops:
        if np.allclose(frac_rot, op["W"], rtol=rtol, atol=atol):
            logger.info(f"Found matching rotation with point group op:\n{op['W']}")
            return True

    logger.info("No matching rotation found")
    return False


def get_cell_from_poscar(poscar: os.PathLike) -> npt.ArrayLike:
    """
    Extract the unit cell from a POSCAR file, including the specified scaling
    """
    with open(poscar) as f:
        poscar_lines = f.read().splitlines()

    scaling = float(poscar_lines[1])
    cell = np.asarray(
        [[float(num) for num in line.split()] for line in poscar_lines[2:5]]
    )

    if scaling < 0:
        desired_volume = -scaling
        unscaled_volume = Cell(cell).volume
        scaling = (desired_volume / unscaled_volume) ** (1 / 3)

    return cell * scaling


def are_in_same_wyckoff_set(letter_1: str, letter_2: str, sgnum: Union[str, int]):
    """
    Given two Wyckoff letters and a space group number, return whether or not they are
    in the same Wyckoff set, meaning that their orbits are related by an operation in
    the normalizer of the space group
    """
    _check_space_group(sgnum)
    with open(os.path.join(DATA_DIR, "wyckoff_sets.json")) as f:
        wyckoff_sets = json.load(f)
    for wyckoff_set in wyckoff_sets[str(sgnum)]:
        if letter_1 in wyckoff_set:
            if letter_2 in wyckoff_set:
                return True
            else:
                return False


def space_group_numbers_are_enantiomorphic(sg_1: int, sg_2: int) -> bool:
    """
    Return whether or not two spacegroups (specified by number) are enantiomorphs of
    each other
    """
    _check_space_group(sg_1)
    _check_space_group(sg_2)
    if sg_1 == sg_2:
        return True
    else:
        enantiomorph_conversion = {
            78: 76,
            95: 91,
            96: 92,
            145: 144,
            153: 151,
            154: 152,
            170: 169,
            172: 171,
            179: 178,
            181: 180,
            213: 212,
        }
        enantiomorph_conversion_2 = {v: k for k, v in enantiomorph_conversion.items()}
        enantiomorph_conversion.update(enantiomorph_conversion_2)
        if enantiomorph_conversion[sg_1] == sg_2:
            return True
        else:
            return False


def get_wyck_pos_xform_under_normalizer(sgnum: Union[int, str]) -> List[List[str]]:
    """
    Get the "Transformed WP" column of the tables at the bottom of the page for each
    space group from https://cryst.ehu.es/cryst/get_set.html
    """
    _check_space_group(sgnum)
    with open(os.path.join(DATA_DIR, "wyck_pos_xform_under_normalizer.json")) as f:
        wyck_pos_xform_under_normalizer = json.load(f)
    return wyck_pos_xform_under_normalizer[str(sgnum)]


def get_bravais_lattice_from_space_group(sgnum: Union[int, str]):
    """
    Get the symbol (e.g. 'cF') of one of the 14 Bravais lattices from the space group
    number
    """
    _check_space_group(sgnum)
    with open(
        os.path.join(DATA_DIR, "space_groups_for_each_bravais_lattice.json")
    ) as f:
        space_groups_for_each_bravais_lattice = json.load(f)
    for bravais_lattice in space_groups_for_each_bravais_lattice:
        if int(sgnum) in space_groups_for_each_bravais_lattice[bravais_lattice]:
            return bravais_lattice
    raise RuntimeError(
        f"Failed to find space group number f{sgnum} in table of lattice symbols"
    )


def get_formal_bravais_lattice_from_space_group(sgnum: Union[int, str]):
    """
    Same as :func:`get_bravais_lattice_from_space_group` except distinguish between "oA"
    and "oC"
    """
    bravais_lattice = get_bravais_lattice_from_space_group(sgnum)
    if bravais_lattice == "oC":
        if int(sgnum) in A_CENTERED_ORTHORHOMBIC_GROUPS:
            return "oA"
        else:
            assert int(sgnum) in C_CENTERED_ORTHORHOMBIC_GROUPS
    return bravais_lattice


def get_primitive_wyckoff_multiplicity(sgnum: Union[int, str], wyckoff: str) -> int:
    """
    Get the multiplicity of a given Wyckoff letter for a primitive cell of the crystal
    """
    _check_space_group(sgnum)
    centering_divisor = CENTERING_DIVISORS[
        get_bravais_lattice_from_space_group(sgnum)[1]
    ]
    with open(os.path.join(DATA_DIR, "wyckoff_multiplicities.json")) as f:
        wyckoff_multiplicities = json.load(f)
    multiplicity_per_primitive_cell = (
        wyckoff_multiplicities[str(sgnum)][wyckoff] / centering_divisor
    )
    # check that multiplicity is an integer
    assert np.isclose(
        multiplicity_per_primitive_cell, round(multiplicity_per_primitive_cell)
    )
    return round(multiplicity_per_primitive_cell)


def get_symbolic_cell_from_formal_bravais_lattice(
    formal_bravais_lattice: str,
) -> Matrix:
    """
    Get the symbolic primitive unit cell as defined in
    http://doi.org/10.1016/j.commatsci.2017.01.017 in terms of the appropriate
    (possibly trivial) subset of the parameters a, b, c, alpha, beta, gamma

    Args:
        formal_bravais_lattice:
            The symbol for the Bravais lattice, e.g "oA". Specifically, "oA" is
            distinguished from "oC", meaning there are 15 possibilities, not just the
            14 Bravais lattices.

    Returns:
        Symbolic 3x3 matrix with the rows being cell vectors. This is in agreement with
        most simulation software, but the transpose of how the ITA defines cell vectors.

    Raises:
        IncorrectCrystallographyException:
            If a nonexistent Bravais lattice is provided
    """
    if formal_bravais_lattice not in FORMAL_BRAVAIS_LATTICES:
        raise IncorrectCrystallographyException(
            f"The provided Bravais lattice type {formal_bravais_lattice} "
            "does not exist."
        )

    a, b, c, alpha, beta, gamma = symbols("a b c alpha beta gamma")

    if formal_bravais_lattice == "aP":
        c_x = c * cos(beta)
        c_y = c * (cos(alpha) - cos(beta) * cos(gamma)) / sin(gamma)
        c_z = sqrt(c**2 - c_x**2 - c_y**2)
        return Matrix([[a, 0, 0], [0, b * cos(gamma), b * sin(gamma)], [c_x, c_y, c_z]])
    elif formal_bravais_lattice == "mP":
        return Matrix([[a, 0, 0], [0, b, 0], [c * cos(beta), 0, c * sin(beta)]])
    elif formal_bravais_lattice == "mC":
        return Matrix(
            [[a / 2, -b / 2, 0], [a / 2, b / 2, 0], [c * cos(beta), 0, c * sin(beta)]]
        )
    elif formal_bravais_lattice == "oP":
        return Matrix([[a, 0, 0], [0, b, 0], [0, 0, c]])
    elif formal_bravais_lattice == "oC":
        return Matrix([[a / 2, -b / 2, 0], [a / 2, b / 2, 0], [0, 0, c]])
    elif formal_bravais_lattice == "oA":
        return Matrix([[a, 0, 0], [0, b / 2, -c / 2], [0, b / 2, c / 2]])
    elif formal_bravais_lattice == "oI":
        return Matrix(
            [[-a / 2, b / 2, c / 2], [a / 2, -b / 2, c / 2], [a / 2, b / 2, -c / 2]]
        )
    elif formal_bravais_lattice == "oF":
        return Matrix([[0, b / 2, c / 2], [a / 2, 0, c / 2], [a / 2, b / 2, 0]])
    elif formal_bravais_lattice == "tP":
        return Matrix([[a, 0, 0], [0, a, 0], [0, 0, c]])
    elif formal_bravais_lattice == "tI":
        return Matrix(
            [[-a / 2, a / 2, c / 2], [a / 2, -a / 2, c / 2], [a / 2, a / 2, -c / 2]]
        )
    elif formal_bravais_lattice == "hP":
        return Matrix(
            [[a / 2, -sqrt(3) * a / 2, 0], [a / 2, sqrt(3) * a / 2, 0], [0, 0, c]]
        )
    elif formal_bravais_lattice == "hR":
        return Matrix(
            [
                [a / 2, -a / (2 * sqrt(3)), c / 3],
                [0, a / sqrt(3), c / 3],
                [-a / 2, -a / (2 * sqrt(3)), c / 3],
            ]
        )
    elif formal_bravais_lattice == "cP":
        return Matrix(
            [
                [a, 0, 0],
                [0, a, 0],
                [0, 0, a],
            ]
        )
    elif formal_bravais_lattice == "cI":
        return Matrix(
            [[-a / 2, a / 2, a / 2], [a / 2, -a / 2, a / 2], [a / 2, a / 2, -a / 2]]
        )
    elif formal_bravais_lattice == "cF":
        return Matrix([[0, a / 2, a / 2], [a / 2, 0, a / 2], [a / 2, a / 2, 0]])
    else:
        assert False


def get_change_of_basis_matrix_to_conventional_cell_from_formal_bravais_lattice(
    formal_bravais_lattice: str,
) -> npt.ArrayLike:
    """
    Get a change of basis matrix **P** as defined in ITA 1.5.1.2, with "old basis"
    being the primitive cell of the provided Bravais lattice, and the "new basis" being
    the conventional cell, i.e. the cell of the primitive lattice of the same crystal
    family. E.g. if ``formal_bravais_lattice="oA"``, then "old basis" is oA, and
    "new basis" is oP. The cell choices are defined in
    http://doi.org/10.1016/j.commatsci.2017.01.017,
    including distinguishing between oA and oC.

    The matrices are given in ITC convention, meaning that they expect to operate on
    column vectors, i.e. The bases are related by the following, where the primed
    symbols indicate the new basis:

    Relationship between basis vectors:
    (**a**', **b**', **c**') = (**a**, **b**, **c**) **P**

    Relationship between fractional coordinates in each basis: **x** = **P** **x**'

    For operating on row vectors, as is often given in simulation software, make sure
    to transpose these relationships appropriately.

    Args:
        formal_bravais_lattice:
            The symbol for the Bravais lattice, e.g "oA". Specifically, "oA" is
            distinguished from "oC", meaning there are 15 possibilities, not just the
            14 Bravais lattices.

    Returns:
        Integral 3x3 matrix representing the change of basis

    Raises:
        IncorrectCrystallographyException:
            If a nonexistent Bravais lattice is provided
    """
    if formal_bravais_lattice not in FORMAL_BRAVAIS_LATTICES:
        raise IncorrectCrystallographyException(
            f"The provided Bravais lattice type {formal_bravais_lattice} "
            "does not exist."
        )

    if formal_bravais_lattice[1] == "P":  # Already primitive
        return np.eye(3)

    corresponding_primitive_lattice = formal_bravais_lattice[0] + "P"

    old_basis = get_symbolic_cell_from_formal_bravais_lattice(
        formal_bravais_lattice
    ).transpose()
    new_basis = get_symbolic_cell_from_formal_bravais_lattice(
        corresponding_primitive_lattice
    ).transpose()

    change_of_basis_matrix = matrix2numpy((old_basis**-1) @ new_basis, dtype=float)

    # matrices should be integral
    assert np.allclose(np.round(change_of_basis_matrix), change_of_basis_matrix)

    return np.round(change_of_basis_matrix)


def change_of_basis_atoms(atoms: Atoms, change_of_basis: npt.ArrayLike) -> Atoms:
    """
    Perform an arbitrary basis change on an ``Atoms`` object, duplicating or cropping
    atoms as needed. A basic check is made that the determinant of ``change_of_basis``
    is compatible with the number of atoms, but this is far from fully determining
    that ``change_of_basis`` is appropriate for the particuar crystal described by
    ``atoms``, which is up to the user.

    TODO: Incorporate :func:`kstest_reduced_distances` into this function

    Args:
        atoms:
            The object to transform
        change_of_basis:
            A change of basis matrix **P** as defined in ITA 1.5.1.2, with ``atoms``
            corresponding to the "old basis" and the returned ``Atoms`` object being
            in the "new basis".

            This matrix should be given in ITC convention, meaning that it expects to
            operate on column vectors, i.e. The bases are related by the following,
            where the primed symbols indicate the new basis:

            Relationship between basis vectors:
            (**a**', **b**', **c**') = (**a**, **b**, **c**) **P**

            Relationship between fractional coordinates in each basis:
            **x** = **P** **x**'

    Returns:
        The transformed ``Atoms`` object, containing the original number of
        atoms mutiplied by the determinant of the change of basis.
    """
    old_cell_column = np.transpose(atoms.cell)
    new_cell_column = old_cell_column @ change_of_basis
    new_cell = np.transpose(new_cell_column)

    # There are surely better ways to do this, but the simplest way I can think of
    # is simply to use ``Atoms.repeat()`` to create a supercell big enough to encase
    # the ``new_cell``, then wrap the atoms back into ``new_cell`` and delete dupes
    repeat = []
    for old_cell_vector in atoms.cell:
        this_repeat = 0
        old_cell_vector_norm = np.linalg.norm(old_cell_vector)
        old_cell_vector_unit = old_cell_vector / old_cell_vector_norm
        # We need to repeat the old vector enough times that it is big enough
        # to cover all possible combinations of projected new vectors
        projections = [
            np.dot(new_cell_vector, old_cell_vector_unit)
            for new_cell_vector in new_cell
        ]
        absmax_projected_sum = 0
        for coeffs in product((-1, 1), repeat=3):
            projected_sum = np.dot(coeffs, projections)
            absmax_projected_sum = max(absmax_projected_sum, abs(projected_sum))
        absmax_projected_sum += 0.1  # pad it a little bit
        this_repeat = ceil(absmax_projected_sum / old_cell_vector_norm)
        repeat.append(this_repeat)

    new_atoms = atoms.repeat(repeat)
    new_atoms.set_cell(new_cell)
    new_atoms.wrap()
    get_duplicate_atoms(new_atoms, delete=True)

    volume_change = np.linalg.det(change_of_basis)
    if not np.isclose(len(atoms) * volume_change, len(new_atoms)):
        raise IncorrectNumAtomsException(
            f"The change in the number of atoms from {len(atoms)} to {len(new_atoms)} "
            f"disagrees with the fractional change in cell volume {volume_change}"
        )

    return new_atoms


def get_possible_primitive_shifts(sgnum: Union[int, str]) -> List[List[float]]:
    """
    Get all unique translation parts of operations in the space group's normalizer that
    don't leave the primitive cell. Given in the primitive basis as defined in
    http://doi.org/10.1016/j.commatsci.2017.01.017

    Args:
        sgnum: space group number
    """
    _check_space_group(sgnum)
    with open(os.path.join(DATA_DIR, "possible_primitive_shifts.json")) as f:
        return json.load(f)[str(sgnum)]


def get_primitive_genpos_ops(sgnum: Union[int, str]) -> List[Dict]:
    """
    Get the matrices and columns of the space group operations in the primitive setting
    as defined in http://doi.org/10.1016/j.commatsci.2017.01.017

    Args:
        sgnum: space group number

    Returns:
        List of dictionaries, with each dictionary containing a matrix 'W' and
        translation 'w' as generally defined in the ITA, but in the primitive setting.
    """
    _check_space_group(sgnum)
    with open(os.path.join(DATA_DIR, "primitive_GENPOS_ops.json")) as f:
        return np.asarray(json.load(f)[str(sgnum)])


def reduce_and_avg(
    atoms: Atoms, repeat: Tuple[int, int, int]
) -> Tuple[Atoms, npt.ArrayLike]:
    """
    TODO: Upgrade :func:`change_of_basis_atoms` to provide the distances
    array, obviating this function

    Function to reduce all atoms to the original unit cell position.

    Args:
        atoms:
            The supercell to reduce
        repeat:
            The number of repeats of each unit cell vector in the
            provided supercell

    Returns:
        * The reduced unit cell
        * An array of displacement vectors. First dimension: index of reference atom
          in reduced cell. Second dimension: index of atom in provided supercell.
          Third dimension: x, y, z
    """
    new_atoms = atoms.copy()

    cell = new_atoms.get_cell()

    # Divide each unit vector by its number of repeats.
    # See
    # https://stackoverflow.com/questions/19602187/numpy-divide-each-row-by-a-vector-element.
    cell = cell / np.array(repeat)[:, None]

    # Decrease size of cell in the atoms object.
    new_atoms.set_cell(cell)
    new_atoms.set_pbc((True, True, True))

    # Set averaging factor
    M = np.prod(repeat)

    # Wrap back the repeated atoms on top of
    # the reference atoms in the original unit cell.
    positions = new_atoms.get_positions(wrap=True)

    number_atoms = len(new_atoms)
    original_number_atoms = number_atoms // M
    assert number_atoms == original_number_atoms * M
    avg_positions_in_prim_cell = np.zeros((original_number_atoms, 3))
    positions_in_prim_cell = np.zeros((number_atoms, 3))

    # Start from end of the atoms
    # because we will remove all atoms except the reference ones.
    for i in reversed(range(number_atoms)):
        if i >= original_number_atoms:
            # Get the distance to the reference atom in the original unit cell with the
            # minimum image convention.
            distance = new_atoms.get_distance(
                i % original_number_atoms, i, mic=True, vector=True
            )
            # Get the position that has the closest distance to
            # the reference atom in the original unit cell.
            position_i = positions[i % original_number_atoms] + distance
            # Remove atom from atoms object.
            new_atoms.pop()
        else:
            # Atom was part of the original unit cell.
            position_i = positions[i]
        # Average
        avg_positions_in_prim_cell[i % original_number_atoms] += position_i / M
        positions_in_prim_cell[i] = position_i

    new_atoms.set_positions(avg_positions_in_prim_cell)

    # Calculate the distances.
    distances = np.empty((original_number_atoms, M, 3))
    for i in range(number_atoms):
        dr, _ = get_distances(
            positions_in_prim_cell[i],
            avg_positions_in_prim_cell[i % original_number_atoms],
            cell=new_atoms.get_cell(),
            pbc=True,
        )
        # dr is a distance matrix, here we only have one distance
        assert dr.shape == (1, 1, 3)
        distances[i % original_number_atoms, i // original_number_atoms] = dr[0][0]

    return new_atoms, distances


def kstest_reduced_distances(
    reduced_distances: npt.ArrayLike,
    significance_level: float = 0.05,
    plot_filename: Optional[str] = None,
    number_bins: Optional[int] = None,
) -> None:
    """
    TODO: Incorporate this into :func:`change_of_basis_atoms`

    Function to test whether the reduced atom positions are normally distributed
    around their average.

    Args:
        reduced_distances:
            Distance array provided by :func:`reduce_and_avg`
        significance_level:
            Significance level for Kolmogorov-Smirnov
        plot_filename:
        number_bins:
            Number of bins for plot

    Raises:
        PeriodExtensionException:
            If a non-normal distribution is detected
    """
    assert len(reduced_distances.shape) == 3
    assert reduced_distances.shape[2] == 3

    if plot_filename is not None:
        if number_bins is None:
            raise ValueError(
                "number_bins must be specified if plot_filename is specified"
            )
        if not plot_filename.endswith(".pdf"):
            raise ValueError(f"{plot_filename} is not a PDF file")
        with PdfPages(plot_filename) as pdf:
            for i in range(reduced_distances.shape[0]):
                fig, axs = plt.subplots(1, 3, figsize=(10.0, 4.0))
                for j in range(reduced_distances.shape[2]):
                    axs[j].hist(reduced_distances[i, :, j], bins=number_bins)
                    axs[j].set_xlabel(f"$x_{j}$")
                axs[0].set_ylabel("Counts")
                fig.suptitle(f"Atom {i}")
                pdf.savefig()
    else:
        if number_bins is not None:
            raise ValueError(
                "number_bins must not be specified if plot_filename is not specified"
            )

    p_values = np.empty((reduced_distances.shape[0], reduced_distances.shape[2]))
    for i in range(reduced_distances.shape[0]):
        atom_distances = reduced_distances[i]

        # Perform PCA on the xyz distribution.
        pca = PCA(n_components=atom_distances.shape[1])
        pca_components = pca.fit_transform(atom_distances)
        assert (
            pca_components.shape == atom_distances.shape == reduced_distances.shape[1:]
        )

        # Test each component with a KS test.
        for j in range(pca_components.shape[1]):
            component = pca_components[:, j]
            component_mean = np.mean(component)
            assert abs(component_mean) < 1.0e-7
            component_std = np.std(component)
            # Normalize component
            normalized_component = (component - component_mean) / component_std
            assert abs(np.mean(normalized_component)) < 1.0e-7
            assert abs(np.std(normalized_component) - 1.0) < 1.0e-7
            res = kstest(normalized_component, "norm")
            p_values[i, j] = res.pvalue

    if np.any(p_values <= significance_level):
        raise PeriodExtensionException(
            "Detected non-normal distribution of reduced atom positions around their "
            f"average (smallest p value {np.min(p_values)})."
        )
    else:
        print(
            "Detected normal distribution or reduced atom positions around their "
            f"average (smallest p value {np.min(p_values)})."
        )


def fit_voigt_tensor_to_cell_and_space_group(
    voigt_input: npt.ArrayLike, cell: npt.ArrayLike, sgnum: Union[int, str]
) -> npt.ArrayLike:
    """
    Given a Cartesian Tensor in voigt form, average it over all the operations in the
    crystal's space group in order to remove violations of the material symmetry due to
    numerical errors. Similar to :meth:`pymatgen.core.tensors.Tensor.fit_to_structure`,
    except the input in output are Voigt, and the symmetry operations are tabulated
    instead of being detected on the fly from a structure.

    The provided tensor and cell must be in the standard primitive
    setting and orientation w.r.t. Cartesian coordinates as defined in
    https://doi.org/10.1016/j.commatsci.2017.01.017

    Args:
        voigt_input:
            Tensor in Voigt form as understood by
            :meth:`pymatgen.core.tensors.Tensor.from_voigt`
        cell:
            The cell of the crystal, with each row being a cartesian vector
            representing a lattice vector
        sgnum:
            Space group number

    Returns:
        Tensor symmetrized w.r.t. operations of the space group
    """
    t = Tensor.from_voigt(voigt_input)

    t_rotated_list = []

    space_group_ops = get_primitive_genpos_ops(sgnum)

    for op in space_group_ops:
        frac_rot = op["W"]
        cart_rot = fractional_to_cartesian_itc_rotation_from_ase_cell(frac_rot, cell)
        cart_rot_op = SymmOp.from_rotation_and_translation(rotation_matrix=cart_rot)
        t_rotated_list.append(t.transform(cart_rot_op))

    t_symmetrized = sum(t_rotated_list) / len(t_rotated_list)

    return t_symmetrized.voigt
